FUNCTION ABAChkSum(cArg)
/*
  -------------------------------------------------------------------
  Programmer..: Al Abbott
  Date........: 10/20/2000
  Notes.......: Used to authenticate USA Bank Transit Nos.
                (Renamed from CkSumABA to ABAChkSum by KMayer)
                Note that the transit number is a nine-digit
                code, so this is the first thing checked ...
  Written for.: Visual dBASE 7.5
  Rev. History: 10/20/2000 -- Original Code
  Calls.......: none
  Usage.......: ABAChkSum( cTransit )
                    or
                :dUFLP:ABAChkSum( cTransit ) // if using source alias
  Example.....: if ABAChkSum( cMyCode )
                   // verified ok processing here
                else
                   // failed verify  processing here
                endif
  Returns.....: logical
  Parameters..: The Bank Transit No. as a character value (cArg)
  ----------------------------------------------------------------- 
*/

   local nCntr,nSum,nMax,nChkNum,cTransit,nDigit,cPos
   store 0 to nCntr,nSum,nMax,nChkNum,nDigit
   store "" to cTransit,cPos
   // added a check for type of argument -- KM
   if type( "cArg" ) # "C"
      cArg := ""+cArg
   endif
   cTransit = cArg.rightTrim().leftTrim()
   // check length:
   if len(cTransit) <> 9
      return false
   endif
   // the last digit is the checksum digit
   nChkNum = val(right(cTransit,1))
   do while nCntr < 8
      nCntr++
      cPos = str(nCntr,1)
      nDigit = val(substr(cTransit,nCntr,1))
      do case
      case cPos $'147'
         nSum += nDigit * 3
      case cPos $'258'
         nSum += nDigit * 7
      case cPos $'36'
         nSum += nDigit * 1
      endcase
   enddo
   if mod(nSum,10) <> 0
      nMax = (int(nSum/10) + 1) * 10
   else
      nMax = nSum
   endif
   if nChkNum <> nMax - nSum
      return false
   endif
return true
//-- EoF: ABAChkSum()
