/*
   Accordian.cc
   Programmer: Ken Mayer
   Date......: May, 2006
   Notes.....: When working with some javascript and web stuff,
               I ran across some spiffy code that was called
               an "accordian" ... the way it works is that
               you have at least two main heading areas,
               and under them is some item or items, such as
               a treeView or a bunch of text, or whatever.
               One of them is always "expanded" and any others
               are always "collapsed", so that the user only
               sees at any one time the contents of one area.

               I worked out a way to do that in dBL, and this
               is my attempt at making it work as a set of custom
               classes.

               This requires two classes: 
                    AccordianMainContainer
                    Accordian

               There needs to be at least two AccordianObj objects,
               otherwise there is no real purpose in using this
               class. These are placed under each other, and
               by default are the same width as the parent class,
               the AccordianMainContainer. This can be as wide
               as needed, and as tall as needed. 


    Usage.....: In the Command Window:
                Set Procedure to :dUFLP:Accordian.cc additive

                Create a form, set the metric to Pixels

                Drag an instance of AccordianMainContainer to 
                  the form.


                Drag an instance of the Accordian class to the 
                   AccordianMainContainer (it has to be on that
                   container). 

                   Place whatever controls you need to on the 
                   'accordian' container. 

                Add as many accordian objects as you wish, one
                under the other, and place controls on them.
                Note that these must be placed in the correct
                sequence (top to bottom -- use the Layout
                menu to set the object order/z-order if needed). 
  
                Change the title text for each as you wish
                by clicking on it and changing the text as
                you normally would.

                Arrange the objects on each container as needed,
                and determine the height of each Accordian container 
                when it is "open". This is important for steps shown
                below ...

                In the form's overridden Open and/or Readmodal methods
                (must be in that code, because it must happen
                before the form opens) -- you might consider
                something like the following:

                function form_open
                   class::Init()
                return FORM::open()

                function form_readModal
                   class::Init()      
                return FORM::readModal()

                function Init
                   // code fired before opening the form

                   // REQUIRED:
                   // Set ExpandedHeight for each accordian object here:
                   private oMain
                   oMain = form.AccordianMainContainer1
                   oMain.Accordian1.ExpandedHeight = 125
                   oMain.Accordian2.ExpandedHeight = 79
                   oMain.Accordian3.ExpandedHeight = 150
          
                   // OPTIONAL: 
                   // set titlebar color for each if you wish
                   // different than default:
                   oMain.Accordian1.TitleBarColor = "Black/White"
                   oMain.Accordian2.TitleBarColor = "Yellow/Red"
                   oMain.Accordian3.TitleBarColor = "Green/Yellow"

                   // REQUIRED: 
                   // add each accordian object to array:
                   oMain.ContArray.add( oMain.Accordian1 )
                   oMain.ContArray.add( oMain.Accordian2 )
                   oMain.ContArray.add( oMain.Accordian3 )
                return
*/
class AccordianMainContainer( oParent ) of Container( oParent ) custom
   with( this )
      onOpen = class::AccordianMainContainer_ONOPEN
      onDesignOpen = class::AMC_onDesignOpen
      width = 300
      height = 350
      borderStyle = 4	// Single
   endwith

   // create array as a custom property:
   this.ContArray = new array()

   function AMC_onDesignOpen( bFromPallette )
      if bFromPallette 
         if form.metric # 6 // pixels
            msgbox( "Form's Metric needs to be pixels, we're setting "+;
                    "it here ... calculations for the containers "+;
                    "require this metric!", "Warning", 48 )
            form.metric := 6 // Pixels
            this.width  := 300
            this.height := 350
         endif
      endif
   return

   function AccordianMainContainer_ONOPEN
      /*
          This code requires modification by the developer
          in that we must assign the names of each Accordian control,
          IN ORDER THAT IT APPEARS ON THE FORM, to the array
          that is created below. 

          The form's overridden OPEN and/or READMODAL event must
          have code to:
            a) set the expanded height for each accordian object
            b) add each to the array needed
          (See heading comments)
      */

      // open the first, collapse the others
      class::OpenAccordian( this.ContArray[1] ) // open first

      // Set mouse event for TitleBar for each:
      for i = 1 to this.ContArray.size
          this.ContArray[i].TitleBar.onLeftMouseUp = Accordian::OpenThis
      next
   return

   function OpenAccordian( oAccord )
      // the tricky thing here is that we have to navigate through
      // the array, open the currently selected object, and 
      // closing the currently open object, moving any containers
      // around as needed ... 
      private nSpacer, oMain, oCont
      nSpacer = 8 // pixels between each container
      // deal with which object is calling this code:
      if this.baseClassName == "FORM"
         oMain = this.AccordianMainContainer1
      elseif this.ClassName == "ACCORDIANMAINCONTAINER"
         oMain = this
      else
         oMain = this.parent.parent  // .parent? may need another level
      endif

      for i = 1 to oMain.ContArray.size
          // is this the container we want to open?
          oCont = oMain.ContArray[i]

          if oCont.Name == oAccord.Name // match
             if not oCont.open
                // set the open property:
                oCont.open := true
                // set the height property:
                oCont.height := oCont.ExpandedHeight
             endif
          else
             // close it:
             oCont.open   := false
             oCont.height := oCont.CollapsedHeight
          endif

          // deal with position:
          if i > 1
             oCont.top := oMain.ContArray[i-1].height+;
                          oMain.ContArray[i-1].top+;
                          nSpacer
          endif
      next
   return

endclass
// ------------------------------------------------------
// Accordian object:
class Accordian( oParent ) of Container( oParent ) custom
   with( this )
      onDesignOpen := class::Accordian_onDesignOpen
      onOpen       := class::Accordian_onOpen
      left         := 0
      width        := this.parent.width
      borderStyle  := 3 // None
      pageNo       := 0
      metric       := 6 // Pixels
   endwith

   // custom properties:
   this.Open            = false
   this.CollapsedHeight = 22 //this.TitleBar.height
   // this.ExpandedHeight must be set
   // when the form opens ... this is the 
   // height of this object when expanded/open
   this.TitleBarColor = "white/navy"

   this.TITLEBAR = new TEXT(this)
   with (this.TITLEBAR)
      height = 22.0
      anchor = 2  // Top
      colorNormal = this.TitleBarColor
      alignHorizontal = 1  // Center
      fontSize = 12.0
      fontBold = true
      text = "Accordian"
      pageno = 0
   endwith

   function Accordian_onDesignOpen( bFromPalette )
      if bFromPalette 
         if this.parent.className # "ACCORDIANMAINCONTAINER"
            msgbox( "This needs to be placed on "+chr(13)+;
                    "the AccordianMainContainer object!",;
                    "ERROR!", 16 )
            return
         else
            this.left  := 0
            this.width := this.parent.width
         endif
      endif
   return

   function Accordian_onOpen
      // force color
      this.TitleBar.colorNormal := this.TitleBarColor
      // force left/width:
      this.left                 := 0
      this.width                := this.parent.width
   return

   function OpenThis
      // called from the TitleBar for the Accordian:
      AccordianMainContainer::OpenAccordian( this.parent )
   return

endclass
