// Metric: 0 - Chars and 6 - Pixels only
/*
    CopyRow.cc

    Programmer..: Ken Mayer
    Date........: April 2, 2002
    Notes.......: This container has two buttons, one of which
                  is used to copy the contents of a row, the
                  other is used to paste that row to a new
                  row. 

                  The idea is to allow, in a data entry situation,
                  something similar to the XDML Set Carry.

                  If, however, all you want to do is copy the row
                  you can use either the kmCarryButton in the 
                  custbutt.cc class of the dUFLP, or the SETCARRY
                  program. Either of these will do what you need.

                  This one has the ability to be useful in a situation
                  such as a parent/child rowset situation, where
                  in you need to copy the contents of the CHILD row,
                  and in another PARENT row add a new child with
                  the same contents.

                  Example: In an awards list tracking system,
                  the people handing out the awards may give
                  the same award to different people on the same
                  day. Rather than re-entering the fields for
                  the award multiple times, you enter it once,
                  click the COPY button of this class, and then 
                  move to the next person the award was given to 
                  (in the parent row), and click the Paste button
                  given here. This will add a new row with all
                  the same information, except where necessary. 
                  For example, there's a linking field in this
                  example for the parent/child relationship,
                  the code should not put the old linking field in,
                  so you would want to add a field list (see examples).

    Rev. History: April 2, 2002 -- Original
                  December, 2002 -- minor modifications in copyrow
                    button to be sure we're not attempting to
                    copy OLE and BINARY field types.

    Usage.......: set procedure to :dUFLP:copyrow.cc additive
                  // drop container onto form
                  // in form designer, using the inspector
                  // set the following in the CopyRowContainer1's
                  // onOpen event:
                  this.RowsetToActOn = form.rowset
                    // or 
                  // this.RowsetToActOn = form.child1.rowset
                  //   or whatever the appropriate rowset reference
                  //   is ...
                  // this is the fieldList if you need it --
                  // list the fieldnames of any fields you 
                  // want copied. If you leave this empty
                  // (or don't enter it at all), all fields
                  // will be copied except AUTOINC fields,
                  // READONLY fields, and/or DBASELOCK fields:
                  this.FieldList = "field1, field2, etc."

                  When run, the copy button will let the user know
                  that the row was copied into memory. The Paste
                  button doesn't really need to because it's obvious.

                  See special notes below if you are using a 
                  complex form with lookupSQL (particularly one
                  with a WHERE clause in the lookupSQL that is 
                  affected by another control ...)

    Special Notes: After a bunch of trial and error, there are some
                   interesting effects in a more complex form.

                   If you have a combobox (or other object and
                   appropriate event) that uses the onSelChange
                   event to set the dataSource for a combobox
                   that is datalinked to a field with a lookupSQL
                   (in the case I ran into, I was using a WHERE
                   clause in the lookupSQL statement), you may find
                   that the results are not returned to you in the
                   fashion you expect.

                   Example:
                      combobox1 has a set of options that is used
                      to set the lookupSQL for combobox2:
                      (this is real code):
                        cRegion = form.notebook1.comboboxregion2.value
                        // change the lookupSQL statement:
                        form.datamodref1.ref.awardlst1.rowset.fields["RoyalCode"].lookupSQL = ;
                            [select royalcode,combined from royals where region="]+;
                            cRegion.rightTrim()+[" order by combined]
                        // restate the datalink:
                        form.notebook1.comboboxroyals.datalink := ;
                             form.notebook1.comboboxroyals.datalink

                   The paste button contained here will allow the first
                   combobox to update properly, no big deal. The second
                   one, however, may not -- even if you have code
                   set to do it. The problem? Unless the first combobox
                   data matches exactly with what you are pasting into
                   that field, the field with the lookup doesn't find
                   a match. The value property is set to empty ...
                   SO, what you have to do is override the PasteRowOnClick
                   button, in this way (again, real code, this works):

                  function PASTEROWBUTTON_onClick
                     // override, but call original code first
                     form.Notebook1.CopyRowContainer1.PasteRowOnClick()
                     // Get the field array
                     aFields = form.Notebook1.CopyRowContainer1.fieldArray
                     // this is the Royalty field in the array:
                     cRoyals = aFields[6]
                     // update the royalty lookupSQL property:
                     class::UpdateRoyaltyCombo()
                     // set the value in this combobox, so we have a match.
                     form.notebook1.comboboxRoyals.value := cRoyals
                     // a bit of extra work, but it DOES work ...
                  return

                  This appears to resolve the problem, in my example,
                  if I save the row, all works properly, and I'm a
                  happy camper ...
*/

class COPYROWCONTAINER(parentObj, name) of CONTAINER(parentObj, name) custom
   with (this)
      transparent = true
      width = 32.2857
      height = 1.3638
      metric = 0	// Chars
      borderStyle = 10	// Etched Out
   endwith

   // Custom Properties
   this.FieldList = null
   this.RowsetToActOn = null

   this.COPYROWBUTTON = new PUSHBUTTON(this)
   with (this.COPYROWBUTTON)
      onClick = class::CopyRowOnClick
      metric = 0	// Chars
      height = 1.0909
      left = 0.14
      top = 0.05
      width = 16.1429
      text = "Copy Current Row"
      pageno = 0
   endwith

   this.PASTEROWBUTTON = new PUSHBUTTON(this)
   with (this.PASTEROWBUTTON)
      onClick = class::PasteRowOnClick
      metric = 0	// Chars
      height = 1.0909
      left = 16.29
      top = 0.05
      width = 15.2857
      text = "Paste Row"
      pageno = 0
   endwith

   function onDesignOpen( lFromPalette )
      if not lFromPalette
         return
      endif
      if form.metric # 0 and form.metric # 6
         msgbox( "Please use either form metric of 0 (Chars) or "+;
                 "6 (Pixels) for this control." )
         return
      endif
      if form.Metric == 6 // Pixels
         this.height                := 30.00
         this.width                 := 226.00
         this.COPYROWBUTTON.height  := 24.00
         this.COPYROWBUTTON.top     := 1
         this.COPYROWBUTTON.width   := 113.00
         this.COPYROWBUTTON.left    := 1
         this.PASTEROWBUTTON.height := 24
         this.PASTEROWBUTTON.top    := 1
         this.PASTEROWBUTTON.width  := 107.00
         this.PASTEROWBUTTON.left   := 114.00
      endif
   return

   function CopyRowOnClick
      // this copies the row to an array:
      // check for errors
      if type( "this.parent.RowsetToActOn" ) # "O"
         msgbox( "The row object reference is not correct.", "CopyRow: ERROR", 48 )
         return
      elseif this.parent.RowsetToActOn.className # "ROWSET"
         msgbox( "The row object reference is not correct.", "CopyRow: ERROR", 48 )
         return
      elseif empty( this.parent.RowsetToActOn )
         msgbox( "There is no rowset object to act on.", "CopyRow: ERROR", 48 )
         return
      endif

      // deal with field list (optional)
      if type( "this.parent.FieldList" ) == "C"
        lList = true
      else
        lList = false
      endif

      // array used to copy current record
      this.parent.FieldArray = new Array()
      // shorten the form.rowset.fields references:
      fFields = this.parent.RowsetToActOn.fields

      // copy current row
      for i = 1 to fFields.size
         lAdd = true
         /*
            except for readonly, autoincrement or lock 
            (_dbaseLock) fields 
             NOTE: _dbaseLock fields don't have
                  type or readOnly properties,
                  so we have to make sure that they
                  exist before we check to see what
                  is in them ... Also we cannot,
                  with the code here, overwrite 
                  autoinc, binary or ole fields.
         */
         if ( type( "fFields[i].readOnly" ) # "U" and ;
             fFields[i].readOnly ) or ;
               ( type ( "fFields[i].type" ) # "U" and ;
               ("AUTOINC" $ fFields[i].type or ;
                "OLE" $ fFields[i].type or ;
                "BINARY" $ fFields[i].type)) or ;
               fFields[i].className == "LOCKFIELD"
            lAdd = false
         endif

         // fields not in FieldList if provided ...
         if lList AND ;
            NOT upper( fFields[i].fieldName ) $ ;
                upper( this.parent.FieldList )
            lAdd = false
         endif

         if not lAdd
            this.parent.FieldArray.add( "* Skip This Field *" )
         else
            /*
               if we have a lookupRowset, then
               we have a lookupSQL statement, and
               we need to ensure that we store the 
               value that needs to be stored, NOT 
               the value that will be returned by 
               the lookupSQL ...
            */
            if fFields[i].lookupRowset # NULL
               this.parent.FieldArray.add( fFields[i].lookupRowset.fields[2].value )
            else
               this.parent.FieldArray.add( fFields[i].value )
            endif
         endif // form.rowset.fields ...
      next

      msgbox( "Row copied, and available to be pasted", "Success" )

   function PasteRowOnClick
      // if you have code that overrides the
      // onClick event (see comments in the header),
      // the object reference changes ... what "this"
      // is, is sort of surprising.
      if this.Name == "PASTEROWBUTTON"
         oThis = this.parent
      else
         oThis = this
      endif

      // error -- no array:
      if type( "oThis.FieldArray" ) == "U" // unknown
         msgbox( "You have not copied a row yet ...", ;
                 "CopyRow: Error", 48 )
         return false
      endif

      // append a new row         
      oThis.RowsetToActOn.beginAppend()
      // shorten the form.rowset.fields references:
      fFields = oThis.RowsetToActOn.fields
            
      // -------------------------------
      // copy data from previous row to
      // new blank row (buffer):
      try
         for i = 1 to fFields.size
            if oThis.FieldArray[i] == "* Skip This Field *"
               loop
            else
               fFields[i].value := oThis.FieldArray[i]
            endif  // skip ...
         next
      catch (Exception E)
         // what do we do if the user tries this on
         // the wrong row (fields may not match), 
         // or some error occurs that is unforseen? 
         // (Bad data? I dunno) We have to let the
         // user know by throwing an error message,
         // and we abandon the new row ...
         msgbox( "Unexpected error: "+e.message+chr(13)+;
                 "while pasting row.", "CopyRow: Error", 48 )
         oThis.RowsetToActOn.abandon()
         return false
      endtry

      // redisplay
      oThis.RowsetToActOn.refreshControls()

      // NOTE: We don't *save* the row here, because
      // the user may wish to abandon it ...
   return true

endclass

