   /* 
     ------------------------------------------------------------------
     Programmer..: Roland Wingerter
     Date........: January 14, 2004
     Notes.......: String class for extracting, drive, path, filename, 
                   file extension from a string  

     Written for.: dBase Plus
     
     Rev. History: August 11, 2004. First version. 
                   January 14, 2005 - simplified driveExists(), no need for an
                      API call, in fact no need for driveExists() - I had not
                      been aware of the built-in function validDrive(). 

     Calls.......: :dUFLP:bdeAlias.cc, 
                   :dUFLP:fileEx.cc, 
                   :dUFLP:getDrive.prg

     Usage.......: set proc to filestring.cc addi
     				    oRef = new filestring([Drive]:\[Path]\[filename.ext]) 

     Example.....: cStr = "C:\dbL Billing Project\tables\employee.dbf"
                   oFS = new filestring(cStr)
                   ? oFS.drivename // "C:"
                   ? oFS.Path //  "C:\dbL Billing Project\tables"
                   ? oFS.fileName // "employee.dbf"
                   ? oFS.fileExtension // ".dbf"

     Returns.....: Object reference with Properties
     ------------------------------------------------------------------
	*/
CLASS filestring( cString ) OF String( cString )
	set procedure to :dUFLP:bdeAlias.cc additive
	set procedure to :dUFLP:fileEx.cc additive
	set procedure to :dUFLP:getDrive.prg additive

   this.string = cString

   this.driveName       = class::getDriveName()
   this.path				= class::getPath()
	this.fileName  		= class::getFileName()
	this.fileExtension  	= class::getFileExtension()
	this.aliasName			= class::getAliasName()
   this.aliasDir		   = class::getDatabaseDir()

   function getDriveName
   	// extract drive name from string
   return this.string.Substring(0,this.String.IndexOf(":")+1)

   function getPath
   	// extract path name from string
   	local cString, cPath
      cString = this.string

      if new fileex().isDir(this.string)
      	// It is a directory
      	cPath = this.string
      else
      	// Take everything up to last backslash
         // Not sure: include backslash or not?
      	cPath = cString.Substring(0,cString.LastIndexOf("\"))
      endif
      
   return cPath

   function getAnyPath
   	// return path from string or database alias
   	local cPath
      cPath = ""
   	if not empty(this.path)
      	cPath = this.path
      elseif not empty(this.aliasDir)
      	cPath = this.aliasDir
      endif
   return cPath

   function getFileName
   	// extract filename from string
   	local cString, cFileName

      cString = this.string

      if cString.indexOf(":") = 0 and cString.lastIndexOf(":") > 1
      	// String starts with alias - extract from last ":"
			cFileName = cString.Substring(cString.LastIndexOf(":")+1, cString.Length)
      elseif empty(this.Path)
         cFileName = cString
      else
         cFilename = cString.Substring(this.Path.Length+1,cString.Length)
      endif
   return cFileName

   function getFileExtension
   	// Extract file extension from string
   	local cString, cExt

      cString = this.string
   	if cString.LastIndexOf(".") > 0
	      cExt = cString.Substring(cString.LastIndexOf(".")+1, cString.Length)
      else
      	cExt = ""
      endif
   return cExt

   function getAliasName
   	// extract alias Name from string
   	local cString, cAlias

      cString = this.string

      if cString.indexOf(":") = 0 and cString.lastIndexOf(":") > 1
      	cAlias = cString.Substring(1, cString.lastIndexOf(":"))
      else
      	cAlias = ""
      endif
   return cAlias

   function getDatabaseDir
   	// Return directory path for database alias
   return new bdeAlias().DatabaseDir( this.AliasName )

   function aliasExists
	   // Check if AliasName exists
   return new bdeAlias().databaseFound( this.AliasName )

   return bIsBDEAlias

   function fileExists
   // Check if file exists
   	local bIsFile, oDB
   	if empty(this.AliasName)
      	bIsFile = file(this.string)
		else
			oDB = new database(this.aliasname)			      	
         oDB.active := true
         bIsFile = oDB.tableExists(this.fileName)
         oDB.active := false
         release oDB
      endif
   return bIsfile

   function getShortPath( nLen )
		if argCount() = 0
      	local nLen
      	nLen = 20
      endif

      local cShortPath 

      cShortPath = this.path

      do while cShortpath.Length > nLen
      	cShortPath = cShortPath.Substring(cShortPath.IndexOf("\")+1, cShortPath.Length)
      enddo
   return cShortPath

   function pathExists
   	// Check if path exists
   return new fileex().isDir(this.path)

   function driveExists
   	// Not needed really, we can use validDrive().
      // Previous version used an API Call for this.
   return validDrive(this.driveName)

   function getDriveType
   return getDriveType(this.driveName)
ENDCLASS
