/*
     SourceAliasPath.prg
     Ken Mayer

     March 1, 2002

     A function that accepts a parameter of a Source Alias,
     and if the alias exists, returns the path. The reason
     this is useful is that Associative Arrays are case-sensitive,
     so using the isKey() method of the _app.SourceAliases array
     may not work, unless you know the case that was used when
     it was created (ex: dUFLP <> duflp <> DUFLP, etc.).

     Usage:
            // optional, depending on where you are
            set procedure to :dUFLP:SourceAliasPath.prg additive
            ? SourceAliasPath( "dUFLP" )
            ? SourceAliasPath( ":dUFLP:" )
            ? SourceAliasPath( ":duflp:MyProg.prg" )

            Note that if a program/procedure file/etc. is 
            included, it will be ignored ... this function
            will simply return the path of the alias itself.
*/
function SourceAliasPath
   parameter cAlias
   local cPath, cCurAlias, cKey, lExists

   if pCount() < 1
      msgbox( "Need an alias to look for.", ;
              "SourceAliasPath::Cannot do it!", 48 )
      return ""
   endif

   cCurAlias = cAlias
   // check for colons, and remove ...
   if at( ":", cCurAlias ) > 0     
      // get alias, and remove : from it
      // if a program/.cc etc. is part of the 
      // parameter, it will be ignored.
      cCurAlias = left( cCurAlias, at(":",cCurAlias,2) -1 )
      cCurAlias = right(cCurAlias,len(cCurAlias) -1 )
   endif

   // search the Source Alias array, but we can't just use IsKey,
   // because Source Aliases are by their nature case sensitive
   cKey = _app.SourceAliases.firstKey
   lExists = false
   for i = 1 to _app.SourceAliases.count()
       if cKey.toLowerCase() == cCurAlias.toLowerCase()
          lExists := true
          cPath = _app.SourceAliases[ cKey ]
          exit
       endif
       cKey = _app.SourceAliases.nextKey( cKey )
   next
   // Problem:
   if not lExists
      msgbox( "Error - Source Alias: "+cCurAlias+" does not exist!",;
              "SourceAliasPath Error", 48 )
      return ""
   endif

return cPath

