// TableDoc.prg
// By Peter Rorlick
// For Visual dBase 7.01
// July 1, 1999
// -- tinkered with by Ken Mayer, July 1, 1999
// -- Modified by Roland Wingerter, August 11, 2004
//    to allow the use of a database alias as parameter 
//
// This program creates a file called
// TableDoc.txt, containing information
// about the tables and indexes in a given directory.
//
// 
// Examples of usage:
//   do TableDoc.prg   // Uses the current directory
//   TableDoc('C:\MyApps\USA\Project1')
//   do tabledoc with ":dbasesamples:" // Documents tables in database
// NOTE: If you have a _dbaselock field in your tables,
// you may want to find a way to deal with it ...
//
// Relies on :dUFLP:FileString.cc

param cPath            // The parameter is optional.

if empty(cPath)
   cPath = ''  // We'll document the tables in the CURRENT directory, 
else
	// added by R.W. Jan, 2005
   set path to :dUFLP:fileString.cc additive
	// Get path from input string using filestring.cc 
   cPath = new filestring(cpath).getAnyPath()+"\"
endif

set safe off
set alte to TableDoc.txt
set alte on
? 'DBF structures, as of '+edate(date())
?

local aFiles, i, nFiles, cFilename
aFiles = new Array()
nFiles = aFiles.Dir( cPath + '*.dbf' )
if nFiles == 0
  msgbox('No tables found')
else
  aFiles.sort(1)  // Sort by file name
  for i = 1 to nFiles
    cFilename = cPath + aFiles[i,1]
    if not aFiles[i,1] = "Backup "
       ? replicate('*',79)
       ListTableStru( cFileName )
       ListIndexInfo( cFileName )
    endif
  next
  use
endif
close alte
set alte to
msgbox('DBF structures and indexes are now listed in TableDoc.TXT')
Modify command TableDoc.txt
return

******************************************
func eDate(d)
// returns a charstring containing an English
// date, e.g. "August 30, 1998"
return cmonth(d)+' '+day(d)+', '+year(d)

******************************************
Func ListTableStru(cTableName)
  ? 
  ? "Table structure for "+cTableName+":"
  local t,i
  t=new tableDef()
  t.tableName = cTableName
  t.load()
  nLengthTotal = 0
  for i = 1 to t.fields.size
      f = t.fields[i]

      cFieldName = trim(f.fieldName) + space( 31 - len( trim( f.fieldName ) ) )
      cType      = trim(f.type) + space( 10 - len( trim( f.type ) ) )
      if f.Type $ "Numeric, Float"
         cLength = transform( f.length, "9999.99" )
      else
         cLength    = transform(f.length, "9999" )+"   "
      endif
      nLengthTotal += f.length
      cDefault   = ""+f.default
      cMaximum   = ""+f.maximum
      cMinimum   = ""+f.minimum
      lRequired  = f.required

      // now to display it all!
      ? "  "+cFieldName+" "+cType+" "+cLength
      if not empty( cDefault)
         ? "   Default: "+cDefault
      endif
      if not empty( cMaximum )
         ? "   Maximum: "+cMaximum
      endif
      if not empty( cMinimum )
         ? "   Minimum: "+cMinimum
      endif
      if lRequired
         ? "   Required"
      endif
  next
  ? "Total Length:                              "+transform(nLengthTotal,"9999")
return

******************************************
Func ListIndexInfo(cTableName)
  ?
  ? 'Indexes for '+cTableName+':'
  local t, i
  t = new tableDef()
  t.tableName = cTableName
  t.load()
  for i = 1 to t.indexes.size
    ?
    with (t.indexes[i])
      ? '  ' + Proper(indexName)
      ? '    ' + expression
      if not empty(ForExpression)
        ? '    For ' + ForExpression
      endif
      if Descending
        ? '    Descending'
      endif
      if Unique
        ? '    Unique'
      endif
      if Type # 0
        ? '    Type: '+Type
      endif
    endWith
  next
return


