// -- File: APPFRAME.cc

/*
   ----------------------------------------------------------------
   Description:

   appFrame.wfm is to demonstrate changing the background color
   of _app.FrameWin

   You can also use appFrameWin to get and/or change the size of
   the application frame window (_app.FrameWin).

   Programmers:

   Vic McClung
   Ken Mayer
   Gary White

    History:

    July 1, 2000 Original version
    July 5, 2000 -- Modified by Ken to make a .CC,
                    and modified appframe.wfm to
                    use the .CC instead ...
    June 4, 2001 -- Fixed problem with 'green' in line 95
    January 7, 2002 -- Updated to add some routines that
                    Gary White created to get and set the 
                    size of the frame window.

    Usage:

    For setting colors:
    You must get the handle of the background window in order
    to manipulated it.  One might naturally assume that would
    be _app.framewin.hwnd, but that isn't the case.  You have
    to use the GetMDIClientHandle method to obtain the handle
    to the background window of the _app.framewin.  Once this
    is accomplished all else is simple.

    Next you retrieve the handle of the current brush used by
    the MDIClient window by calling GetClassLong api function.
    This handle must be saved and restore upon exit of your
    application.  After this is accomplished, you must create
    your own brush every time you change the background color.
    After you are finished with each brush you must use the
    api call DeleteObject to destroy the brush and reclaim the
    resources used by it.

    Once you have created a brush and retrieved the handle to
    that brush, use the SetClassLong api call to change colors
    of the background.  Then merely invalidate the MDIClient
    window and it will be repainted by windows.

    For getting/setting position, you use _app.framewin.hwnd.

    set procedure to appframewin.cc additive
       or
    set procedure to :dUFLP:appframewin.cc additive
    _app.FramewinColor = new AppFrameWinColor()
    _app.FramewinColor.appOpen()
    // turn off the shell
    shell(.f., .f.)

    // to get user defined color:
    _app.FramewinColor.SelectColor()
    // or, to set your own color (you have to
    // use the colors returned from the getColor()
    // function)
    _app.FramewinColor.SetColor( cRed, cGreen, cBlue )

    // when application is closed, you should always
    // use this:
    _app.FramewinColor.appClose()

   // to use the code to get or change the position of the framewindow:
   set procedure to appFrameWin.cc additive
   local f
   f = new AppFramewinColor()

   // show the current position and size
   ? f.top
   ? f.left
   ? f.height
   ? f.width

   // if the frame window is moved, or resized by the user, you can
   // determine the current position and size by calling getPos():
   f.getPos()
   ? f.top
   ? f.left
   ? f.height
   ? f.width

   // Set the frame window to 640x480 in the upper left corner of the screen
   f.setPos( 0, 0, 640, 480)
   return

*/

class AppFramewinColor
   // extern declarations for WINAPI32 calls used in this program

   // prototypes used for getting/setting colors
   extern chandle  CreateBrushIndirect(cptr) gdi32
   extern clong    SetClassLongA(chandle, cint, clong) user32
   extern clong    GetClassLongA(chandle, cint) user32
   extern cint     GetClassNameA(chandle, cptr, cint) User32
   extern chandle  GetWindow(chandle, cint) User32
   extern clogical DeleteObject(chandle) gdi32
   extern clogical GetClientRect(chandle, cptr) user32
   extern clogical InvalidateRect(chandle, cptr, clogical) user32

   // prototypes used for getting/setting size/position
   extern clogical SystemParametersInfo( ;
      cuint, cuint, cptr, cuint)user32 ;
      from "SystemParametersInfoA"
   extern clogical MoveWindow( ;
      chandle, cint, cint, cint, cint, clogical) user32
   extern clogical GetWindowRect(chandle, cptr) user32

   // defines for WINAPI32 Calls
   #define GW_HWNDNEXT  2
   #define GW_CHILD     5
   #define HWND_BOTTOM  1
   #define GCL_HBRBACKGROUND   (-10)

   // added for Gary's additional code
   #define SPI_GETWORKAREA 48

   this.left = 0
   this.top = 0
   this.width = 640
   this.height = 480
   this.rect = new RECT()
   this.GetPos()

   function SelectColor
      // Function to select a color, and then
      // deal with painting the background of
      // the _app.framewin
      local cLOGBRUSH, cColor, red, green, blue
      cColor = getColor()
      if empty(cColor) // user canceled dialog
         return
      endif
      red = substr(cColor, 1, at(',', cColor) - 1)
      green = substr(cColor, at(',', cColor, 1) + 1, ;
              at(',', cColor, 2) - at(',', cColor, 1) - 1)
      blue = substr(cColor, at(',', cColor, 2)+ 1)
      class::SetColor( red, green, blue )
   return

   function SetColor( Red, Green, Blue )
      cLOGBRUSH = class::num2chr32(0)+class::num2chr32(class::rgb(val(red), ;
                  val(green), val(blue)))+class::num2chr32(0)
      DeleteObject(_app.hBrush)
      _app.hBrush = CreateBrushIndirect(cLOGBRUSH)
      SetClassLongA(_app.hMDIClient, GCL_HBRBACKGROUND, _app.hBrush)
      class::repaint()
   return

   function appOpen
      local cLOGBRUSH, oConvert
      // determine the handle of the MDIClient Window
      _app.hMDIClient = class::GetMDIClientHandle()
      // rgb macro here determines the color
      // if you want to change it just run the getcolor() function and use the
      // 3 numbers returned from it to plug into the rgb macro below.  For example
      // blue will be 0, 0, 255
      // red will be 255, 0, 0
      // green will be 0, 255, 0
      // create the LOGBRUSH structure containing the RGB color
      // this structure is 0000XXXX0000 where XXXX is a 32 bit number for the RGB Color
      cLOGBRUSH = class::num2chr32(0)+class::num2chr32(class::rgb(9, 9, 225))+class::num2chr32(0)
      // create the logical brush and return the handle
      _app.hBrush = CreateBrushIndirect(cLOGBRUSH)
      // retrieve the current brush handle and save it
      // restore it upon application termination
      _app.hOldBrush = GetClassLongA(_app.hMDIClient, GCL_HBRBACKGROUND)
      // set the background color
      SetClassLongA(_app.hMDIClient, GCL_HBRBACKGROUND, _app.hBrush)
      class::repaint()

   function appClose
      // you must do this at close of application!!!!
      // reset the background color to orginal
      SetClassLongA(_app.hMDIClient, -10, _app.hOldBrush)
      // delete the brush you created
      DeleteObject(_app.hBrush)

   function RGB(r, g, b)
      // convert the 3 numbers returned by getColor into
      // a 32 bit ColorRef for use by Windows
   return (r + g*256 + b*65536)

   function Num2Chr32(nVal)
      // this function changes a numerical value to a 32 bit
      // number in the format used by 'C' in windows, 32 bit
      // unsigned long integer
      local b1, b2, b3, b3, dw
      dw = new String(space(2)) // 4 bytes
      b1 =           BitAnd(nVal, htoi("000000FF"))
      b2 = BitRShift(BitAnd(nVal, htoi("0000FF00")), 8)
      b3 = BitRShift(BitAnd(nVal, htoi("00FF0000")), 16)
      b4 = BitRShift(BitAnd(nVal, htoi("FF000000")), 24)
      if b4 < 0
         b4 = 256 + b4
      endif
      dw.SetByte(0, b1)
      dw.SetByte(1, b2)
      dw.SetByte(2, b3)
      dw.SetByte(3, b4)
   return dw

   Procedure SBCSZToDBCS(s)
      // convert a Single Byte Character String
      // with zero terminator to a Double Byte
      // Character string for use in dBASE 32 bit
      local cTemp, i, nLen, cChar
      nLen = (len(s)*2)-1
      cTemp = ''
      for i = 0 to nLen
         cChar = chr(s.getByte(i))
         if cChar == chr(0)
            *remove the 'C' null string terminator
            exit
         endif
         cTemp += cChar
      next
   return cTemp

   Procedure GetMDIClientHandle
      // this procedure returns the handle of the
      // MDIClient window, the "real" background of
      // _app.FrameWin
      local cBuffer, hMDIClient, cName, oString
      hMDIClient = GetWindow(_app.FrameWin.hWND, GW_CHILD)
      * search child windows until MDICLIENT window class found
      do while hMDIClient > 0
            cBuffer = space(81)
            GetClassNameA(hMDIClient, cBuffer, 80)
            cName = class::SBCSZToDBCS(cBuffer)
            if upper(ltrim(rtrim(cName))) == "MDICLIENT"
               exit  && MDIClient window has been found
            endif
            hMDIClient = GetWindow(hMDIClient, GW_HWNDNEXT)
      enddo
   return hMDIClient

   function repaint()
      // invalidate the MDIClient window rectangle
      // and Windows will repaint it
      local cRect
      cRect = space(8)
      GetClientRect(_app.hMDIClient, cRect)
      InvalidateRect(_app.hMDIClient, cRect, true)
   return

   function GetPos
   	// populates the left, top, width and height properties
      local b
      b = false
      if GetWindowRect(_app.framewin.hWnd, this.rect)
         this.left := this.rect.getLeft()
         this.top := this.rect.getTop()
         this.width := this.rect.getWidth()
         this.height := this.rect.getHeight()
         b := true
      endif
   return b

   function SetPos(nLeft, nTop, nWidth, nHeight)
   	// sets the _app.framewin position and size
      local r, b
      b = false
      r = new RECT()
      if SystemParametersInfo(SPI_GETWORKAREA, 0, r, 0)

         // insure new size does not exceed screen size
         nHeight := min(r.getHeight(), nHeight)
         nWidth := min(r.getWidth(), nWidth)

         // insure we don't position off screen
         nTop := min(nTop, r.getHeight() - nHeight)
         nTop := max(nTop, 0)
         nLeft := min(nLeft, r.getWidth() - nWidth)
         nLeft := max(nLeft, 0)

      endif
      b := MoveWindow(_app.framewin.hWnd, nLeft, nTop, nWidth, nHeight, true)
   return b

endclass
// end of class: appFramewin

class RECT of struct
	// A RECT structure, used in API calls is composed of 4 long 
   // integers, for a total of 16 bytes. Each character in dB2K
   // is composed of 2 bytes, so we need an 8 character string
   // to simulate the 16 byte structure.

   this.string := space(8)

   // The left value in the structure is bytes 0 thru 3
   function getLeft
   return int(this.getLong(0))

   function setLeft(n)
   return this.setLong(0, n)

   // The top value in the structure is bytes 4 thru 7
   function getTop
   return int(this.getLong(4))

   function setTop(n)
   return this.setLong(4, n)
                       
   // The right value in the structure is bytes 8 thru 11
   function getRight
   return int(this.getLong(8))

   function setRight(n)
   return this.setLong(8, n)

   // The bottom value in the structure is bytes 12 thru 15
   function getBottom
   return int(this.getLong(12))

   function setBottom(n)
   return this.setLong(12, n)

   function getWidth
   return int(this.getRight() - this.getLeft())

   function setWidth(n)
   return this.setRight(this.getLeft() + n)

   function getHeight
   return int(this.getBottom() - this.getTop())

   function setHeight(n)
   return this.setBottom(this.getTop() + n)

endclass
// end of class: Rect

class struct of string
	// Base structure class for use with API calls needing a
   // POINT, or RECT structure

   function getLong(nIndex)
   	// converts the 4 bytes, starting with byte # nIndex
      // into a long integer
      local n
      n = this.getByte(nIndex) + ;
         bitlshift(this.getByte(nIndex + 1), 8) + ;
         bitlshift(this.getByte(nIndex + 2), 16) + ;
         bitlshift(this.getByte(nIndex + 3), 24)
      if bitset(n, 31)
         n := - bitnot(n)
      endif
   return n

   function setLong(nIndex, nValue)
   	// converts a long integer into its 4 bytes and stores them
      // beginning with byte # nIndex
      local bRet
      try
         bRet = true
         this.setByte(nIndex, bitand(nValue, 0xFF))
         this.setByte(nIndex + 1, bitrshift(bitand(nValue, 0xFF00), 8))
         this.setByte(nIndex + 2, bitrshift(bitand(nValue, 0xFF0000), 16))
         this.setByte(nIndex + 3, bitrshift(bitand(nValue, 0xFF000000), 24))
      catch(exception e)
         bRet = false
      endtry
   return bRet

   function release
      this.string = ''
      release object this
   return

endclass
// end of class: struct

// ------------------------------
// - end of file: appFrameWin.cc
// ------------------------------


