/*
   COUNTQRY.CC
   A query with an overriden count method that is more accurate
   than the one that is part of the rowset.

   This is based on code posted by Bowen Moursund or
   Peter Rorlick in the Visual dBASE newsgroups, Feb. 19, 1999

   Works like a normal query/rowset, except for the addition of
   the over-ridden method.

*/

class CountQry of query() custom

   procedure execute
      super::execute() // execute the query, produce the rowset
      this.rowset.count := {|| CountQry::Count() }
   return

   function count
      // the overridden method of the rowset --
      // during this code, "this" refers to the
      // rowset, not the query, as the rowset is
      // what will be calling this code:
      local b, nCount, lSaveNotify

      // don't notify any active controls that are 
      // datalinked of what we are doing:
      lSaveNotify         = this.notifyControls
      this.notifyControls := false

      // save current location in the table:
      b      = this.bookMark()
      // set the counter:
      nCount = 0

      // loop through the table:
      this.first()
      do while not this.endOfSet
         nCount++
         this.next()
      enddo

      // back to where we were:
      this.goto( b )

      // make it possible to notify the controls again:
      this.notifyControls := lSaveNotify

   return nCount

endclass
