/*
     HTMLCLAS.CC
     Programmer..: Ken Mayer and Gary White
     Date........: 09/30/1998
                   10/3/1998 -- Gary White enhanced this
                                quite a bit ... including 
                                making this able to create
                                the HTML file rather than 
                                having to do it as a separate
                                task and call this class.
                   March 8, 1999 -- minor enhancements
                                to use String Object's
                                methods in some cases,
                                added some new routines
                                for Anchor Tags

                   August 28, 2002 -- Minor modifications
                                to deal with XHTML standards,
                                all tags should be lower case,
                                and almost all have an ending
                                tag ... in most cases this
                                is always dealt with here, except
                                for the paragraph tag, which 
                                should have a </p> tag.

                   January 24, 2003 -- Added EMailObscure (and
                                Obscure()) method to deal with
                                making life difficult for 
                                spammers.

     Notes (there are a lot -- HTML is a big subject):
     This class is designed to help a developer generate 
     HTML output in their code ... the example in usage (below)
     assumes you wish to create an HTML file and uses
     low-level file handling to do so -- this is VERY 
     fast. There is a sample program in the dUFLP library
     that tests nearly all possible tag/attribute combinations
     that are used here, called TESTHTML.PRG -- if you
     want to see it in action, run that program and 
     then look at the resulting HTML file (TESTIT.HTM) 
     in your browser ...

     *** Note that the STRING class actually contains a lot
     of the functionality given here -- there are many methods
     of this class that produce HTML ... this can be useful
     if all you want is to generate some HTML in a standard
     dBASE form or report ... some of that is used in these
     methods (for example the "bold" function calls
     the string bold() method, which places the bold
     <B> and </B> tags around the text ...)

     *******************************************************
     This class assumes you know something about
     HTML, and does little error checking for attributes
     passed to the various tags -- in some cases a tag may
     be streamed out without the incorrect attribute, otherwise
     we assume you know what you're doing ...
     *******************************************************

     This class assumes HTML level 3.2, and does not currently 
     include all HTML features.

     Note that this class does not include FORM elements
     (entryfields and such) -- that's another area entirely,
     and if you want to use those with Visual dBASE, you should
     be checking into the Web Wizards and such that ship
     with the product.

     This class does not include some of the variations, such as
     <EM> for "Emphasize" which is equivalant to Italic (<I>)
     as it seemed redundant, and there are other tags
     that the authors didn't feel like throwing in here, either.
     I suppose a purist would put them all in -- if you feel
     like being a purist and "finishing" the job, go for it. 
     <G> (And there are also some attributes I didn't include --
     if you really need them you could add them ...)

     COLORS:
        Anywhere you can specify colors in HTML you can do it
        in one of two ways. The first, and "older" method
        is to use the hexadecimal RGB values. These are
        tricky to work with -- and this class assumes
        that if you want to use them, that you will pass
        the "$" that is necessary for HTML to interpret them
        properly:
           Example:  red = "$FF0000"  (the first two characters
                                       after the $ are the
                                       hex for red, FF means
                                       "all the red you got"
                                       and lesser values mean
                                       less red. 00 in the Green
                                       positon means no green,
                                       and 00 in the Blue position
                                       means no blue ... hence
                                       black would be 000000 (no
                                       color) and white would
                                       be FFFFFF (all colors))
                     or RED           Text word "red" gets 
                                      translated into the hex
                                      value internally by
                                      the HTML browser.

           If you enter a character color value that the browser
           doesn't know how to interpret, you may get some
           interesting results. In addition, many folk
           recommend that you use the "safe color palette" --
           this is something you should look up if you 
           wish to "be safe" ... some browsers interpret
           colors outside of that pallette ... strangely.
           (For a resource for the color safe palettes,
           I use "Web by Design" by Molly E. Holzchlag)

     Usage (example):
        set procedure to htmlclas.cc additive
        h = new htmlPage(<cFileName>)
        // if cFileName is not supplied you will 
        // be prompted for a filename
        // Also note that HTMLStart() is implicit here ...
        h.HeadBegin()
        h.Title( "Some Title" )
        h.CommentStart()
        h.write( "   Generated by Visual dBASE 7 code" )
        h.write( "   Created by: Ken Mayer and Gary White" )
        h.write( "   Date: "+date() )
        h.CommentEnd()
        h.HeadEnd()
        h.BodyBegin("","GREEN","YELLOW")

        h.centerStart()
	  	  h.HeadingStart(1)
        h.write( "This is a heading" )
        h.HeadingEnd(1)
        h.centerEnd()

        //    or:

        // the advantage here is you can build the output
        // string programmatically and then do the h.write()
        // once you are sure you're done ...

        cOutString = "This is a heading"
        h.centerStart()
        h.HeadingStart(1)
        h.write( cOutString )
        h.HeadingEnd(1)
        h.centerEnd()

        // whatever else -- loop through a table,
        // etc.

        // you might want to use ParseAnsi() to 
        // handle special characters ( ansi > 128 )
        cCheck = fMyQuery.rowset.fields["somefield"].value
        h.write( ParseAnsi( cCheck ) )

        // don't forget these:
        h.CloseFile() // will add bodyEnd() and 
                      // HTMLEnd() calls automatically
                      // (thanks to Gary)


     Methods:
       Note, as many HTML tags require an end tag, there are
       many "matching" methods here -- one to start, one to
       end. The idea is that these will simply return the
       appropriate tag. In some cases, however, there are
       many options for the individual HTML tags, so there
       are parameters that may be passed -- and the value
       returned should include the correct HTML attributes ...

       Details on various HTML used here and explanations
       of the attributes (parameters) are based on 
       Queue's "USING HTML (Special Edition)" which is based
       on HTML 3.0 specification.

       ****************************************************
       As these are all methods, they must have a set of
       parenthesis after them (htmlclas.methodName() )
       or they will not work properly, even if there are
       no parameters passed to the method
       ****************************************************

         Document Stuff:
         ---------------
         HTMLStart()        <html>  // automatically called when
                                    // instantiating the html object
         HTMLEnd()          </html> // automatic if closeFile() 
                                    // method is used
                                    

         HeadStart()        <head>
         HeadEnd()          </head>

         Title("Some Text") <title>Some Text</title>

         BodyStart()        <BODY BACKGROUND=... BGCOLOR=... etc.>
            Optional parameters:
            BACKGROUND = a URL or image name for a background
                         image -- JPG or GIF
            BGCOLOR    = either a color *name* or $RRGGBB hexadecimal
                         string for the Background color of the 
                         HTML document
            TEXT       = color as above for text on document
            LINK       = hyperlink color as above
            VLINK      = 'visited' links in the document -- color
                          as above
            ALINK      = "active" hyperlinks as above (link
                         the mouse is *clicked* on).

            If you do not wish to use a parameter, but wish
            to use the ones *after* one you are skipping, make
            sure you put double quotes:  

                h.BodyBegin( "". "BLACK", "RED" )

                will produce an html document with no background
                image, but a black background with red text
                as the standard text color 

         BodyEnd()         </BODY>   // automatically called if
                                     // closeFile() method is used

         Comment( cText )  <!-- content of cText -->
         CommentStart()    <!-- 
         CommentEnd()      -->

         -----------
         Formatting:
         -----------
         Address( cText )   <ADDRESS>cText</ADDRESS>
         AddressStart       <ADDRESS>
         AddressEnd         </ADDRESS>

         Bold( cText )      <B>cText</B>
         BoldStart          <B>
         BoldEnd            </B>

         // Blockquote indents text ...
         BlockQuoteStart    <BLOCKQUOTE>
         BlockQuoteEnd      </BLOCKQUOTE>

         // break to next line with no blank line
         Break( clear )     <BR> or <BR CLEAR=...>
             Clear is useful to clear an image for the
             text ... if the image is on the left, the
             right or both sides -- basically
             it means to not display any text
             if an image is on the left, right or
             both sides -- once clear of the image,
             text will start to display. (It keeps you
             from having to issue a bunch of paragraph or
             break tags, which will evaluate differently
             from one browser to another ...)
           CLEAR = LEFT, RIGHT, ALL

         // Center the text -- you can center
         // a bunch of text by calling CenterStart()
         // and then outputting the text. If you want
         // breaks, use the Break() method ... make
         // sure you end with the CenterEnd() method:
         //   CenterStart()
         //   h.write( "A bunch o' text" )
         //   h.break() 
         //   h.write("more text" )
         //   h.CenterEnd()
         // if you want to center a single line, just
         // use the Center() method ...
         Center( cText )    <CENTER>cText</CENTER>
         CenterStart        <CENTER>
         CenterEnd          </CENTER>

         // Heading
         Heading( nSize, cText )
                            <Hn>cText</Hn>
         HeadingBegin(n)    <Hn>
            n is a value from 1 to 6 (1 is largest)
         HeadingEnd(n)      </Hn>

         // Italic
         Italic( cText )    <I>cText</I>
         ItalicStart        <I>
         ItalicEnd          </I>

         // Listing ignores HTML and does NO formatting
         ListingStart       <LISTING>
         ListingEnd         </LISTING>

         // don't break text within these tags
         // This is useful for really long lines
         // that you want to continue to be really
         // long lines -- otherwise HTML will
         // break them as needed to fit on the 
         // screen.
         NoBreak( cText )   <NOBR>cText</NOBR>
         NoBreakStart       <NOBR>
         NoBreakEnd         </NOBR>

         // Paragraph - break to next line but
         // insert a blank line first
         Paragraph( ALIGN )  <P> or <P ALIGN=...>
           ALIGN = LEFT, RIGHT, CENTER
              if you use this, you need paragraphEnd,
              otherwise you don't ...
         ParagraphEnd        </P>

         // PRE does no formatting, but uses
         // current font (retains breaks in text ...)
         PreStart           <PRE>
         PreEnd             </PRE>

         // horizontal rule
         Rule( parameters )
           optional:
           WIDTH = absolute in pixels or percentage
           SIZE  = height
           NOSHADE = put anything here, and this will
                     not appear as three-d
           ALIGN = left, right, center

         // Strikeout
         Strikeout( cText )  <S>cText</S>
         StrikeOutStart      <S>
         StrikeOutEnd        </S>

         // Subscript
         Subscript( cText )  <SUB>cText</SUB>
         SubscriptStart      <SUB>
         SubscriptEnd        </SUB>

         // Superscript
         Superscript( cText ) <SUP>cText</SUP>
         SuperscriptStart     <SUP>
         SuperscriptEnd       </SUP>

         // Typewriter Text -- uses mono-spaced font defined
         // in browser, requires <BR> for line breaks ...
         TTStart            <TT>
         TTEnd              </TT>

         // Underline:
         Underline( cText ) <U>cText</U>
         UnderlineStart     <U>
         UnderlineEnd       </U>

         -----
         Font:
         -----
         BaseFont( nSize )
           nSize = 1 through 7, 3 is default, 7 is largest,
                   1 is smallest font (opposite of Hn tag)
                   Do *not* use relative font sizes -- they
                   do not work and the +/- sign gets
                   ignored ...

         FontStart( nSize, cColor )
           SIZE = 1 through 7 or
                  relative (+1, -2 ...) -- relative needs
                  to be in quotes -- use single
                  quotes, i.e.,  FontStart('+1')  
                  You should probably just include whatever
                  parameter in quotes here, as HTML will
                  recognize it.
                  (default font size is 3)
           COLOR = RGB hex values or color name (i.e., "$FF0000" or
                                                       "RED")
         FontEnd            </FONT>

         -------
         Tables:
         -------
         // A caption goes either above or below the
         // table -- Captions are optional 
         CaptionStart( cAlign )
            cAlign = TOP or BOTTOM 
         CaptionEnd

         // begin a table
         TableStart			 <TABLE parameters>
            BORDER numeric, line weight
            CELLSPACING numeric - spacing in pixels 
                        between adjoining cells
            CELLPADDING numeric - vertical spacing 
                        between adjacent cells
            WIDTH       numeric - number of pixels
                                  or percentage (i.e., 50%)
                                  of individual cells

          // End the table - don't forget this!
         TableEnd           </TABLE>

         // If you want headings on the top or
         // left of a table, you can do it yourself
         // or use this -- it automatically 
         // bolds the text -- use this in place
         // of the TableCellStart and TableCellEnd 
         // methods (TableHeadings are optional)
         TableHeadStart     <TH parameters>
            ALIGN = LEFT, RIGHT, CENTER
            VALIGN = TOP, MIDDLE, BOTTOM
            COLSPAN = numeric -- number of columns
                      heading will span
            ROWSPAN = numeric -- number of rows heading
                      will span
            WIDTH   = numeric - number of pixels or
                      percentage ...
            NOWRAP == put anything here and 
                      contents of cell will not wrap

         TableHeadEnd       </TH>

         // You need one of these for each table row
         // and must have a TableRowEnd()
         TableRowStart      <TR parameters>
            ALIGN = LEFT, RIGHT, CENTER
            VALIGN = TOP, MIDDLE, BOTTOM
         TableRowEnd        </TR>

         // Actual table data, one cell at a time ...
         TableCellStart     <TD parameters>
            ALIGN   = LEFT, RIGHT, CENTER
            VALIGN  = TOP, MIDDLE, BOTTOM
            COLSPAN = numeric -- number of columns
                      heading will span
            ROWSPAN = numeric -- number of rows heading
                      will span
            WIDTH   = numeric - number of pixels or
                      percentage ...
            NOWRAP  = put anything here and 
                      contents of cell will not wrap

         TableCellEnd       </TD>

         // or ...
        	TableCell( cContent, <TD parameters as above> )


         // Example to create a simple table:
         // CaptionStart("TOP")+"Test Table"+CaptionEnd()
         // TableStart()
         // TableRowStart()
         //    TableHeadStart()
         //			Write("Column1")
         //    TableHeadEnd()
         //    TableHeadStart()
         //			Write("Column2")
         //		TableHeadEnd()
         // TableRowEnd()
         // TableRowStart()
         //     TableCell("Some data")
         //     TableCell("Some more data")
         // TableRowEnd()
         // TableEnd()


         ------
         Image:
         ------
         Image              <IMG SRC='filename.jpg' otherparms>
           URL (required) is path to file/image
           optional:
           ALT   = text to display if graphic support
                   is not available or other problems occur
           (leaving out ISMAP -- image maps are not
            something I felt like tackling here)
           ALIGN = Alignment of text on same
                   line as current inline graphic:
                   TOP, MIDDLE, BOTTOM, LEFT, RIGHT,
                   TEXTTOP, ABSMIDDLE, BASELINE, ABSBOTTOM
           HSPACE = value in pixels between image and text
                    horizontal
           VSPACE = as above but vertical
           WIDTH  = width of image in pixels
           HEIGHT = same as above but height
           BORDER = border for inline image -- pixels
           (leaving out LOWSRC ...)

           Ex: htmlobj.Image( "..\images\someimage.jpg", "An Image",;
                             "TOP", 5, 5 )

         -------
         Anchor:
         -------
         This is done in two sections -- one is "encapsulated",
         which means you can simply specify directly the
         complete anchor tag (begin/end tags). The other
         is in parts which allows some flexibility, such
         as using an image for a link, or a combination
         of an image and text ...

         Encapsulated:
         Link               <A HREF='cURL'>cText</A>
           cText is the text to display as a link
           cURL is the URL linking to
         Anchor             <A NAME = 'cName'>cText</A>
           cText is the text to place in the anchor
           cName is the name associated with it
         MailTo             <A HREF='mailto:cMailTo>cText</A>
           cText is the text to place between the tags
           cMailTo is the email address

         Broken out into parts (this allows flexibility
         for what is inside the tags ...):
         LinkStart          <A HREF='cURL'>
           cURL is the URL ...
         LinkEnd            </A>
         LinkName           <A NAME='name'>
           Should have a call to LinkEnd() after it.
         LinkMailStart      <A HREF='mailto:address@somewhere'>
           Must have a call to LinkEnd() after it, but
           should have text and/or image immediately 
           after this one ...

         // Link examples:
         // ENCAPSULATED:
         // Link( "someone's web site", "http://somewhere.com/mypage.htm" )
         //    Outputs: <A HREF='http://somewhere.com/mypage.htm'>someone's website</A>
         // 
         // Anchor( "Fred's Information", "Fred" )
         //    Outputs: <A NAME='Fred'>Fred's Information</A>
         //
         // MailTo( "Email Me!", "someone@somewhere.com" )
         //    Outputs: <A HREF='mailto:someone@somewhere.com'>Email Me!</A>  
         // 
         // LinkStart( "http://somewhere.com/mypage.htm" )+;
         //    "Someone's web site"+LinkEnd()
         // 
         // IN PARTS:
         // LinkName( "Fred" )
         //		Write( "Fred's Information" )
         //	LinkEnd()
         //
         // (To use link to Fred in same page:)
         // LinkStart( "#Fred" )
         //		Write( "To Fred Link" )
         //	LinkEnd()
         // (To use link to Fred in a different page: )
         // LinkStart( "http://somewhere.com/mypage.htm#Fred" )
         //   Write("Link to Fred at Someone's web site" )
         //	LinkEnd()
         //
         // Mail:
         // LinkMailStart( "mymail@myip.com" )
         // 	Write("EMail me!")
         //	LinkEnd()
         //
         // NOTE that you can use an image for a link:
         // LinkMailStart( "mymail@myip.com" )
         //    Image( "myimage.jpg", "Mail Me", "", 5, 5,"","" , 0 )
         // LinkEnd()
         // The image tag that is output will give the name
         // of the image, an "Alt" tag, sets horizontal
         // and vertical spacing to 5 pixels (so text doesn't
         // come closer than 5 pixels to the image) and sets
         // the border to zero, so no border appears around
         // the image on the page ...

         ------
         Lists:
         ------
         // To see examples of these, run the TESTHTML.PRG -- 
         // there's too many permutations ...

         // Note that while you must have an end tag for the list,
         // that the ListItem is considered to be anything
         // up to the next ListItem or the end tag ...

         OrderedListStart   <OL TYPE=. START=n>
           (An "Ordered List" is one that has some sort of
            numbering system to show the level ... the default
            is numeric as noted below )

            TYPE parameter is:
               A. defines number characters as capital letters
               a. defines as lower-case
               I. capital roman numerals
               i. lower-case roman numerals
               1. numbers (default)
            START = value list starts at ... (numeric)

         OrderedListEnd     </OL>

         UnorderedListStart <UL TYPE=...>
           (An unordered list has a bullet instead of a numeric
            or alpha character to show the level or point in the 
            list

            For whatever reason, Internet Exploder 4 does not
            recognize the variations here:)

            TYPE parameter is:
               CIRCLE
               SQUARE
               DISC 

         UnorderedListEnd   </UL>

         ListItem           <LI TYPE=... VALUE=n>
            (These are the actual bulleted items in the list
             whether using OrderedList or UnorderedList ...)

            TYPE parameter is (based on above):
               CIRCLE
               SQUARE
               DISC
               A
               a
               I
               i
               1
            VALUE is numeric (sequential value)

         --------------------------
         Other Interesting Methods:
         --------------------------
         EMailObscure( cEMail, cTitle )
            This outputs a small JavaScript routine that
            literally makes it impossible for most spambots
            to find an email address in the HTML, as the
            address never appears with the html "mailto:"
            option, nor does it appear with an @ symbol
            in the source ... This calls Obscure() method.

            Assuming an email address of: Tester@test.com,
            and a 'title' of "EMail Tester", you should see
            JavaScript like the following, in the HTML
            source:

            <script language="JavaScript" type="text/javascript">
            <!--
               var first = 'ma';
               var second = 'il';
               var third = 'to:';
               var address = '&#84;&#101;&#115;&#116;&#101;&#114;';
               var domain = '&#116;&#101;&#115;&#116;&#46;&#99;&#111;&#109;';
               document.write('<a href="');
               document.write(first+second+third);
               document.write(address);
               document.write('&#64;');
               document.write(domain);
               document.write('" title="EMail Tester">');
               document.write('EMail Tester<\/a>');
            // -->
            </script>

         Obscure( cString )
            Returns a string where each character is converted
            to HTML &#nnnn; values ...

         Symbol( cSymb )
            Where cSymb (defined below) is a specific
            symbol you need to output in HTML, but
            doesn't appear the way you expect it to ...

             "NBSP"         // non-breaking space
             "LT" OR "<"    // < (less than)
             "GT" OR ">"    // > (greater than)
             "AMP" OR "&"   // & (ampersand)
             "?"            // ? (question mark)
             "I!"           //  (inverted excl)
             "I?"           //  (inverted question)
             "CENT"         //  (cent)
             "#"            //  (pound sterling)
             "CURR"         //  (general currency)
             "YEN"          //  (Yen)
             "|"            // vertical bar
             "COPY"         //  Copyright
             "REG"          //  (Registered Trademark)

          You could add more -- see documentation
          for this routine ...


         ParseAnsi( cValue )
            Scans through a character string, character
            by character, and converts characters that
            have ANSI values 128 and up to their HTML 
            equivalent so that they display correctly. 
            This is useful for those who need to display 
            names (for example) with, say,  ... this will
            not display properly if you just throw it out
            there -- HTML won't know what to do with 
            it. However, this routine will convert it
            to the appropriate:  "&#252;"  character 
            combination which HTML *does* understand.
            This routine was originally designed for characters
            => 192, for some SCA stuff I was doing, but 
            characters that are > 128 will have problems,
            so it was modified to handle those ...

            ** Caveat: This routine will slow down
               processing a table, but it does make
               things easier ...

*/

class HTMLPage( cFileName )
   // The following code added by Gary White:
   private cSafety

   if empty( cFileName )
	   cSafety = set("SAFETY")
	   set safety on
   	cFileName = putFile( "Enter a name for the HTML Page",;
		   "*.htm*", "htm" )
      set safety &cSafety
      release cSafety
      if empty( cFileName )
      	return
      endif
   endif
   this.outFile = new file()
   try
   	this.outFile.create( cFileName )
   catch ( exception e )
      if type( "cMessage" ) == "U" // undefined
         cMessage = e.message
      endif
	   msgbox( cMessage, "Error", 16 )
   endtry
   this.write([<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0//EN">])
   this.FileName = cFileName
   this.fileIsOpen = true
   this.bodyStarted = false
   this.htmlStart()

   function release
      if this.fileIsOpen
         this.closeFile()
      endif
   release object this

   function closeFile
   	if this.bodyStarted
      	this.bodyEnd()
      endif
      this.htmlEnd()
      this.fileIsOpen = false
   return this.outFile.close()

   function write( cText )
   return this.outfile.puts( cText )

   // -----------------------------------------------------
   // Document tags:
   // -----------------------------------------------------
   function HTMLStart // should be at beginning of document
   return this.write( "<html>" )

   function HTMLEnd   // should be at end of document
   return this.write( "</html>" )

   function HeadStart
   return this.write( "<head>" )

   function HeadEnd
   return this.write( "</head>" )

   function Title( cTitle ) // should be inside head and /head
   return this.write( "<title>" + cTitle + "</title>" )

   // body should come after the headEnd() call:
   function BodyStart( cImage, cBGColor, cText, cLink, cVLink, cALink )
      cReturn = "<body"
      if pCount() > 0 and not empty( cImage )
         cReturn += [ background="]+cImage+["]
      endif
      if pCount() > 1 and not empty( cBGColor )
         cReturn += [ bgcolor="]+cBGColor+["]
      endif
      if pCount() > 2 and not empty( cText )
         cReturn += [ text="]+cText+["]
      endif
      if pCount() > 3 and not empty( cLink )
         cReturn += [ link="]+cLink+["]
      endif
      if pCount() > 4 and not empty( cVLink )
         cReturn += [ vlink="]+cVLink+["]
      endif
      if pCount() > 5 and not empty( cALink )
         cReturn += [ alink="]+cALink+["]
      endif
      cReturn += ">"
      this.bodyStarted = true
   return this.write( cReturn )

   // should be right before the HTMLEnd() call -- this 
   // is the next-to-last tag in a standard HTML document
   function BodyEnd
	   this.bodyStarted = false
   return this.write( "</body>" )

   // comments can be placed anywhere, but always need the commentEnd()
   function Comment( cText )
   return this.CommentStart() + this.write( cText ) + this.CommentEnd()

   function CommentStart
   return this.write( "<!-- " )

   function CommentEnd
   return this.write( " -->" )

   // -----------------------------------------------------
   // Formatting tags:
   // Note that some of these allow you to call the
   // start/end tags explicitly as well as simply
   // pass a text parameter, such as the first --
   // with "ADDRESS" you can simply pass an "Address"
   // as text, and it will be output with the start/end
   // tags for you, so you don't have to call these
   // as separate calls.
   // -----------------------------------------------------
   function Address( cText )
   return this.AddressStart() + this.write( cText ) + ;
   	this.AddressEnd()

   function AddressStart
   return this.write( "<address>" )

   function AddressEnd
   return this.write( "</address>" )

   function Block( cText )
   return this.BlockStart() + this.write( cText ) + this.BlockEnd()

   function BlockStart
   return this.write( "<blockquote>" )

   function BlockEnd
   return this.write( "</blockquote>" )

   function Bold( cText )
   return this.write( cText.bold() )

   function BoldStart
   return this.write( "<b>" )

   function BoldEnd
   return this.write( "</b>" )

   function Break( cClear )
      cReturn = "<br"
      if pCount() > 0 and not empty( cClear )
         if NOT upper( cClear ) $ "LEFT, RIGHT, ALL"
            msgbox( "<br clear='"+cClear+"'> is not valid HTML",;
                    "Storing plain tag" )
         else
            cReturn += [ clear="] + cClear +["]
         endif
      endif
      cReturn += " />"
   return this.write( cReturn )

   function Center( cText )
   return this.centerStart() + this.write( cText ) + ;
   	this.centerEnd()

   function CenterStart
   return this.write( "<center>" )

   function CenterEnd
   return this.write( "</center>" )

   function Heading( nSize, cText )
   return this.headingStart( nSize ) + this.write( cText ) + ;
          this.HeadingEnd( nSize )

   function HeadingStart(n) // minimal checking here
      do case
         case n < 1 
              cReturn = "<h1>"
         case n > 6
              cReturn = "<h6>"
         otherwise
              cReturn = "<h" + n + ">"
      endcase
   return this.write( cReturn )

   function HeadingEnd(n)
      do case
         case n < 1 
              cReturn = "</h1>"
         case n > 6
              cReturn = "</h6>"
         otherwise
              cReturn = "</h" + n + ">"
      endcase
   return this.write( cReturn )

   function Italic( cText )
   return this.write( cText.italics() ) 

   function ItalicStart
   return this.write( "<i>" )

   function ItalicEnd
   return this.write( "</i>" )

   function ListingStart
   return this.write( "<listing>" )

   function ListingEnd
   return this.write( "</listing>" )

   function NoBreak( cText )
   return this.NoBreakStart() + this.write( cText  ) + ;
   	this.NoBreakEnd()

   function NoBreakStart
   return this.write( "<nobr>" )

   function NoBreakEnd
   return this.write( "</nobr>" )

   function Paragraph( cAlign )
      cReturn = "<p"
      if pCount() > 0 and not empty( cAlign )
         if NOT upper( cAlign ) $ "LEFT, RIGHT, CENTER"
            msgbox( "<p align="+cClear+"> is not valid HTML",;
                    "Storing plain tag" )
         else
            cReturn += [ align="]+cAlign+["]
         endif
      endif
      cReturn += ">"
   return this.write( cReturn )

   function ParagraphEnd
   return this.write( "</p>" )

   function PreStart
   return this.write( "<pre>" )

   function PreEnd
   return this.write( "</pre>" )

   // Horizontal Rule:
   function Rule( nWidth, nSize, NoShade, cAlign )
      cReturn = "<hr"
      if pCount() > 0 and not empty( nWidth )
         cReturn += [ width="]+nWidth+["]
      endif
      if pCount() > 1 and not empty( nSize )
         cReturn += [ size="]+nSize+["]
      endif
      if pCount() > 2 and not empty( NoShade )
         cReturn += " noshade"
      endif
      if pCount() > 3 and not empty( cAlign )
         if NOT upper( cAlign ) $ "LEFT, RIGHT, CENTER"
            msgbox( cReturn + [ align="]+cAlign+["> is not valid HTML],;
                    "Storing without bad ALIGN attribute" )
         else      
            cReturn += [ align="]+cAlign+["]
         endif
      endif
      cReturn += ">"
   return this.write( cReturn )

   function Strikeout( cText )
   return this.write( cText.strike() )

   function StrikeoutStart
   return this.write( "<strike>" )

   function StrikeoutEnd
   return this.write( "</strike>" )

   function Subscript( cText )
   return this.write( cText.sub() )

   function SubscriptStart
   return this.write( "<sub>" )

   function SubscriptEnd
   return this.write( "</sub>" )

   function Superscript( cText )
   return this.write( cText.sup() )

   function SuperscriptStart
   return this.write( "<sup>" )

   function SuperscriptEnd
   return this.write( "</sup>" )

   function TT( cText )
   return this.write( cText.fixed() )

   function TTStart
   return this.write( "<tt>" )

   function TTEnd
   return this.write( "</tt>" )

   function Underline( cText )
   return this.UnderlineStart() + this.write( cText ) + ;
		this.UnderlineEnd()

   function UnderlineStart
   return this.write( "<u>" )

   function UnderlineEnd
   return this.write( "</u>" )

   // -----------------------------------------------------
   // Font tags:
   // -----------------------------------------------------
   function BaseFont( nSize )
   return this.write( [<basefont size="]+nSize+[">] )

   function FontStart( cSize, cColor )
      cReturn = "<font"
      if pCount() > 0 and not empty( cSize )
         do case
            case cSize < 1
                 cReturn += [ size="1"]
            case cSize > 7
                 cReturn += [ size="7"]
            otherwise
                 cReturn += [ size="]+cSize+["]
         endcase
      endif
      if pCount() > 1 and not empty( cColor )
         cReturn += [ color="] + cColor + ["]
      endif
      cReturn += ">"
   return this.write( cReturn )

   function FontEnd
   return this.write( "</font>" )

   // -----------------------------------------------------
   // Table tags:
   // -----------------------------------------------------
   function CaptionStart( cAlign )
      cReturn = "<caption"
      if pCount() > 0 and not empty( cAlign )
         if NOT upper( cAlign ) $ "TOP, BOTTOM"
            msgbox( cReturn + [ align="]+cAlign+["> is not valid HTML],;
                    "Storing without bad ALIGN attribute" )
         else
            cReturn += [ align="]+cAlign+["]
         endif
      endif
      cReturn +=">"
   return this.write( cReturn )

   function CaptionEnd
   return this.write( "</caption>" )

   function TableStart( nBorder, nCellSpacing, nCellPadding, nWidth )
      cReturn = "<table"
      if pCount() > 0 and not empty( nBorder )
         if nBorder < 0
            msgbox( "Border of '0' is invalid -- not storing this attribute.",;
                    "Bad attribute" )
         else
            cReturn += [ border="]+nBorder+["]
         endif
      endif
      if pCount() > 1 and not empty( nCellSpacing )
         cReturn += [ cellspacing="]+nCellSpacing+["]
      endif
      if pCount() > 2 and not empty( nCellPadding )
         cReturn += [ cellpadding="]+nCellPadding+["]
      endif
      if pCount() > 3 and not empty( nWidth )
         cReturn += [ width="]+nWidth+["]
      endif
      cReturn += ">"
   return this.write( cReturn )

   function TableEnd
   return this.write( "</table>" )

   function TableHeadStart( cAlign, cVAlign, nColSpan, nRowSpan, nWidth, NoWrap )
      cReturn = "<th"
      cLegal = "LEFT|RIGHT|CENTER"
      if pCount() > 0 and not empty( cAlign )
         if not upper( cAlign ) $ "LEFT, RIGHT, CENTER"
            msgbox( cReturn + [ align="]+cAlign+["> is not valid HTML],;
                    "Storing without bad ALIGN attribute" )
         else
            cReturn += [ align="]+cAlign+["]
         endif
      endif             
      if pCount() > 1 and not empty( cVAlign )
         if not upper( cVAlign ) $ "TOP, MIDDLE, BOTTOM"
            msgbox( cReturn + [ valign="]+cVAlign+["> is not valid HTML],;
                    "Storing without bad VALIGN attribute" )
         else
            cReturn += [ valign="]+cAlign+["]
         endif
      endif
      if pCount() > 2 and not empty( nColSpan )
         cReturn += [ colspan="]+nColSpan+["]
      endif
      if pCount() > 3 and not empty( nRowSpan )
         cReturn += [ rowspan="]+nRowSpan+["]
      endif
      if pCount() > 4 and not empty( nWidth )
         cReturn += [ width="]+nWidth+["]
      endif
      if pCount() > 5 and not empty( NoWrap )
         cReturn += [ nowrap]
      endif
      cReturn += ">"
   return this.write( cReturn )

   function TableHeadEnd
   return this.write( "</th>" )

   function TableRowStart( cAlign, cVAlign )
      cReturn = "<tr"
      if pCount() > 0 and not empty( cAlign )
         if not upper( cAlign ) $ "LEFT, RIGHT, CENTER"
            msgbox( cReturn + [ align="]+cAlign+["> is not valid HTML],;
                    "Storing without bad ALIGN attribute" )
         else
            cReturn += [ align="]+cAlign+["]
         endif
      endif
      if pCount() > 1 and not empty( cVAlign )
         if not upper( cVAlign ) $ "TOP, MIDDLE, BOTTOM"
            msgbox( cReturn + [ valign="]+cVAlign+["> is not valid HTML],;
                    "Storing without bad VALIGN attribute" )
         else
            cReturn += [ valign="]+cVAlign+["]
         endif
      endif
      cReturn += ">"
   return this.write( cReturn )

   function TableCell( cText, cAlign, cVAlign, nColSpan, nRowSpan, nWidth, NoWrap )
   return this.TableCellStart( cAlign, cVAlign, nColSpan, nRowSpan, nWidth, NoWrap ) +;
   	this.write( cText ) + this.TableCellEnd()

   function TableRowEnd
   return this.write( "</tr>" )

   function TableCellStart( cAlign, cVAlign, nColSpan, nRowSpan, nWidth, NoWrap )
      cReturn = "<td"
      if pCount() > 0 and not empty( cAlign )
         if not upper( cAlign ) $ "LEFT, RIGHT, CENTER"
            msgbox( cReturn + [ align="]+cAlign+["> is not valid HTML],;
                    "Storing without bad ALIGN attribute" )
         else
            cReturn += [ align="]+cAlign+["]
         endif
      endif
      if pCount() > 1 and not empty( cVAlign )
         if not upper( cVAlign ) $ "TOP, MIDDLE, BOTTOM"
            msgbox( cReturn + [ valign="]+cVAlign+["> is not valid HTML],;
                    "Storing without bad VALIGN attribute" )
         else
            cReturn += [ valign="]+cVAlign+["]
         endif
      endif
      if pCount() > 2 and not empty( nColSpan )
         cReturn += [ colspan="]+nColSpan+["]
      endif
      if pCount() > 3 and not empty( nRowSpan )
         cReturn += [ rowspan="]+nRowSpan+["]
      endif
      if pCount() > 4 and not empty( nWidth )
         cReturn += [ width="]+nWidth+["]
      endif
      if pCount() > 5 and not empty( NoWrap )
         cReturn += " nowrap"
      endif
      cReturn += ">"
   return this.write( cReturn )

   function TableCellEnd
   return this.write( "</td>" )

   // -----------------------------------------------------
   // Image:
   // -----------------------------------------------------
   function Image( cURL, cAlt, cAlign, nHSpace, nVSpace, nWidth, nHeight, nBorder )
      cReturn = [<img src="] + cURL + ["]
      if pCount() > 1 and not empty( cAlt )
         cReturn += [ alt="]+cAlt+["]
      endif
      if pCount() > 2 and not empty( cAlign )
//      <IMG ALIGN= left|right|top|texttop|middle|absmiddle|baseline|bottom|absbottom>
         if not upper( cAlign ) $ "LEFT|RIGHT|TOP|TEXTTOP|MIDDLE|ABSMIDDLE|BASELINE|BOTTOM|ABSBOTTOM"
            msgbox( cReturn + [ align="]+cAlign+["> is not valid HTML],;
                    "Storing without bad ALIGN attribute" )
         else
            cReturn += [ align="]+cAlign+["]
         endif
      endif
      if pCount() > 3 and not empty( nHSpace )
         cReturn += [ hspace="]+nHSpace+["]
      endif
      if pCount() > 4 and not empty( nVSpace )
         cReturn += [ vspace="]+nVSpace+["]
      endif
      if pCount() > 5 and not empty( nWidth )
         cReturn += [ width="]+nWidth+["]
      endif
      if pCount() > 6 and not empty( nHeight )
         cReturn += [ height="]+nHeight+["]
      endif
      if pCount() > 7 and not empty( nBorder )
         cReturn += [ border="]+nBorder+["]
      endif
      cReturn += " />"
   return this.write( cReturn )

   // -----------------------------------------------------
   // Anchors:
   // -----------------------------------------------------

   // Encapsulated:
   function Link( cText, cURL )
   return this.write( cText.link( cURL ) )

   function Anchor( cText, cName )
   return this.write( cText.anchor( cName ) )

   function MailTo( cText, cMailTo )
   return this.write( [<a href="mailto:]+cMailTo+[">] + cText + [</a>] )

   // In Parts:
   function LinkStart( cURL )
   return this.write( [<a href="]+cURL+[">] )

   function LinkEnd()
   return this.write( [</a>] )

   function LinkName( cName )
   return this.write( [<a name="]+cName+[">] )

   function MailToStart( cMailTo )
   return this.write( [<a href="mailto:]+cMailTo+[">] )

   // -----------------------------------------------------
   // Lists
   // -----------------------------------------------------
   // Ordered List:
   function OrderedListStart( cType, nStart )
      cReturn = "<ol"
      if pCount() > 0 and not empty( cType )
         if not upper( cType ) $ "A,I,1"
            msgbox( cReturn + [ type="]+cType+["> is not valid HTML],;
                    "Storing without bad TYPE attribute" )
         else
            cReturn += [ type="]+cType+["]
         endif
      endif
      if pCount() > 1 and not empty( nStart )
         if nStart < 1
            msgbox( nStart+" is invalid start value",;
                    "Storing without bad START attribute" )
         else
            cReturn += [ start="]+nStart+["]
         endif
      endif
      cReturn += ">"
   return this.write( cReturn )

   function OrderedListEnd
   return this.write( "</ol>" )

   // Ordered List:
   function UnOrderedListStart( cType )
     cReturn = "<ul"
     if pCount() > 0 and not empty( cType )
        if not upper( cType ) $ "CIRCLE, SQUARE, DISC"
           msgbox( cReturn + [ type="]+cType+["> is not valid HTML],;
                   "Storing without bad TYPE attribute" )
        else
           cReturn += [ type="]+cType+["]
        endif
      endif
     cReturn += ">"
   return this.write( cReturn )

   function UnOrderedListEnd
   return this.write( "</ul>" )

   // ListItem:
   function ListItem( cType, nValue )
      cReturn = "<li"
      if pCount() > 0 and not empty( cType )
         if not upper( cType ) $ "A, I, 1, CIRCLE, SQUARE, DISC"
            msgbox( cReturn + [ type="]+cType+["> is not valid HTML],;
                    "Storing without bad TYPE attribute" )
         else
            cReturn += [ type="]+cType+["]
         endif
      endif
      if pCount() > 1 and not empty( nValue )
         if nValue < 1
            msgbox( nValue+" is invalid value",;
                    "Storing without bad VALUE attribute" )
         else
            cReturn += [ value="]+nValue+["]
         endif
      endif
      cReturn += ">"
   return this.write( cReturn )

   // -----------------------------------------------------
   // Other useful methods:
   // -----------------------------------------------------
   function EMailObscure( cEMail, cTitle )
      // This generates a javascript routine
      // that will obscure an email address for 
      // a "Mailto:", to avoid letting spammers get 
      // people's email from websites. 
      // Requires two parameters:
      //   cEMail = the email address
      //   cTitle = the text displayed for the address
      //            i.e., "EMail Me"
      // This calls OBSCURE ...
      local cLeft, cRight
      if pCount() == 0 or pCount() == 1
         msgbox( "Need two parameters: cEMail and cTitle!",;
                 "HTMLCLAS.CC::EMailObscure() error", 16 )
         return
      endif
      this.outFile.puts( [<script language="JavaScript" type="text/javascript">] )
      this.outFile.puts( [<!--] )
      this.outFile.puts( [var first = 'ma';] )
      this.outFile.puts( [	var second = 'il';] )
      this.outFile.puts( [	var third = 'to:';] )
      // get left side of email address (everything to left of '@' symbol
      // and convert to special characters:
      cLeft = left( cEMail, at("@", cEmail ) -1 )
      // call routine to convert each character:
      cLeft = this.Obscure( cLeft )
      this.outFile.puts( [	var address = ']+cLeft+[';] )
      // get the right side of the '@' symbol:
      cRight = right( cEMail, len( cEMail ) - at( "@", cEMail ) )
      cRight = this.Obscure( cRight )
      this.outFile.puts( [	var domain = ']+cRight+[';] )
      this.outFile.puts( [document.write('<a href="');] )
      this.outFile.puts( [document.write(first+second+third);] )
      this.outFile.puts( [document.write(address);] )
      this.outFile.puts( [document.write('&#64;');] )
      this.outFile.puts( [document.write(domain);] )
      this.outFile.puts( [document.write('" title="]+cTitle+[">');] )
      this.outFile.puts( [document.write(']+cTitle+[<\/a>');] )
      this.outFile.puts( [/]+[/ -->] )
      this.outFile.puts( [</script>] )

   function Obscure( cString )
      // return HTML values for characters
      local cReturn, nPos, cChar
      cReturn = "" // blank
      for nPos = 1 to len( cString )
          cChar = substr( cString, nPos, 1 ) // 1 char at a time
          cReturn += "&#"+asc( cChar )+";"
      next
   return cReturn

   function Symbol( cSymb )
      cSymb = cSymb.toUpperCase()
      cReturn = ""
      do case
         case cSymb == "NBSP" // non-breaking space
              cReturn := "&#160;"
         case cSymb == "LT" OR ;
              cSymb == "<"    // < (less than)
              cReturn := "&lt;"
         case cSymb == "GT" OR;
              cSymb == ">"     // > (greater than)
              cReturn := "&gt;"
         case cSymb == "AMP" OR;
              cSymb == "&"    // & (ampersand)
              cReturn := "&#62;"
         case cSymb == "?"    // ? (question mark)
              cReturn := "&#34;"
         case cSymb == "I!"   //  (inverted excl)
              cReturn := "&#161;"
         case cSymb == "I?"   //  (inverted question)
              cReturn := "&#191;"
         case cSymb == "CENT"  //  (cent)
              cReturn := "&#162;"
         case cSymb == "#"     //  (pound sterling)
              cReturn := "&#163;"
         case cSymb == "CURR"  //  (general currency)
              cReturn := "&#164;"
         case cSymb == "YEN"   //  (Yen)
              cReturn := "&#165;"
         case cSymb == "|"     // vertical bar
              cReturn := "&#166;"
         case cSymb == "COPY"  //  Copyright
              cReturn := "&#167;"
         case cSymb == "REG"   //  (Registered Trademark)
              cReturn := "&#174;"
         // there's more if you feel like looking
         // them up ... this web site is where I got
         // them from originally:
         // http://www.uni-passau.de/~ramsch/iso8859-1.htm
      endcase
   return this.write( cReturn )

   function ParseAnsi( cValue )
      private cValue   // so we don't change
      cField = cValue  // *original* value ...

      // Going to have to loop through, find the number of
      // ansi characters, and then loop that number of times
      // through this thing:
      nAnsiCount = 0
      for nPos = 1 to cField.Length
          cChar = substr( cField, nPos, 1 )
          if asc( cChar ) > 128
             nAnsiCount++
          endif
      next // nPos

      // If we have any, loop through and find them, and
      // perform a substitution of "&#" and the numeric value
      if nAnsiCount > 0
         nLength = cField.Length + ( nAnsiCount * 5 )
         for nPos = 1 to nLength
             cChar = substr( cField, nPos, 1 )
             if asc( cChar ) > 128 
                // time to convert it:
                cLeft  = left( cField, nPos - 1 )
                cRight = right( cField, cField.length - ( nPos ) )
                cField = cLeft + "&#"+ asc(cChar) + ";" + cRight
             endif // asc( cChar ) > 128
         next // nPos
      endif
   return this.write( ( cField ) )


endclass

// ---------------------------------------------------------------
// End of Class: HTMLCLAS.CC
// ---------------------------------------------------------------
