/* 
   TableRestructure : Version 1.13

   File			: modistru.cc			
   Author		: Marc Van den Berghen
   Date 			: 14-09-03
   Purpose		: Designed to make changes in tables without using the 
                 Tabledesigner. 
                 It can be used to change the fieldtype from long 
                 to autoinc f.i. or to change the length of the 
                 decimal part in numerics ...
   Depends on  : api_structs.cc
   Usage			: local mstru
                 mstru = new modistruclass()
                 mstru.cTableName = Name of the table
                 mstru.oDataBase =  if a database-object is used, you can pass it here
                 * if the databaseobject is null, the standarddatabase 
                 * is used
                 mstru.cFieldName= Name of the field to be changed(or added)
                 mstru.NewFieldName = if not empty, this is the new fieldname
                 mstru.NewFieldType	= new type of field (for change and add)
                 mstru.NewFieldLength = <ilength>
                 mstru.NewFieldDecimal	= <idecimals>   
                 mstru.restructure()     // make a change in the field-structure
                 mstru.ErrorNumber =  Result of the restructuring .
                 mstru.ErrorString =  Error-Designation
                 mstru.release()
                 mstru.addfield()		//  add a field
                 mstru.deletefield()	// delete a field	
                 You can change more than one field-property at 
                 a time, just set the apropriate members of the class
                 for the different Fieldtypes see the defines below.

//============================================================================
//                              Field Types
//============================================================================

Paradox types (Physical)

#define  CHAR      		   0x101        // Alpha    (string)
#define  NUMERIC 		      0x102        // Numeric
#define  MONEY 		      0x103        // Money
#define  DATE 			      0x104        // Date
#define  SHORT 		      0x105        // Short
#define  MEMO  		      0x106        // Text Memo       (blob)
#define  BINARYBLOB 			0x107        // Binary data     (blob)
#define  FMTMEMO  		   0x108        // Formatted text  (blob)
#define  OLEBLOB  		   0x109        // OLE object      (blob)
#define  GRAPHIC  		   0x10A        // Graphics object (blob)
#define  LONG  		      0x10B        // Long
#define  TIME		         0x10C        // Time
#define  DATETIME  		   0x10D        // Time Stamp
#define  BOOL  		      0x10E        // Logical
#define  AUTOINC   		   0x10F        // Auto increment (long)
#define  BYTES   			   0x110        // Fixed number of bytes
#define  BCD   		      0x111        // BCD (32 digits)

xBASE types (Physical)

#define  CHAR   		 	   0x201        // Char string
#define  NUMERIC     	   0x202        // Number
#define  MEMO  		      0x203        // Memo          (blob)
#define  BOOL  		      0x204        // Logical
#define  DATE  		      0x205        // Date
#define  FLOAT   		      0x206        // Float
#define  DBLOCK   	      0x207        // Logical type is LOCKINFO
#define  OLE			      0x208        // OLE object    (blob)
#define  BINARY            0x209        // Binary data   (blob)
#define  BYTES  		      0x20A        // Only for TEMPORARY tables
#define  LONG	  			   0x20B			 // Long
#define	TIMESTAMP			0x20C			 // Timestamp
#define	DOUBLE				0x20D			 // Double
#define	AUTOINC				0x20E			 // Autoincrement


   Thanks to 	: Ken Mayer, who started the whole tabledesigner-task
                 Michael Nuwer who pointed out that dbiDoRestructure 
                 exists 
                 Stuart Sidders who asked to change fieldwith in 
                 runtime	                             	                            

   History		: Version 1.0		14-09-03  	Marc VdB
                 first beta to be beaten on in the NG  <g>

                 Version 1.01		16-09-03
                 - changed the field translation from logical to 
                   physical, so that subtypes are no concern anymore 
                 - added Error-Strings (thanks to Michael Nuwer)
                 - added properties to make a method call without 
                   parameters
                 - the fieldtypes are assigned by name. (Thanks to Ken)
   
	       		  Version 1.11	20-09-03
                 - addfield  (see method-header for explication)
                 - deletefield ( see method-header)
                 - init() is now called from the methods, so the developper 
                   has not to bother with it

                 Version 1.12 21-09-03
                 - getdblocklength-method.

                 Version 1.13  2012/12/01   Mervyn Bick
                 - corrected problem whereby spurious characters were left
                   in fieldname if newfieldname is shorter than oldfieldname



   *** Important *** Important *** Important *** Important *****
   *                                                           *
   *       DON'T USE LIVETABLES TO TEST THIS CLASS             *
   *  IF YOU PASS WRONG PARAMETERS, YOU CAN LOOSE YOUR DATA !  *
   *                                                           *
   *************************************************************
*/


// Driver Types

#define szPARADOX       "PARADOX"		 //  only those two are supported
#define szDBASE         "DBASE"

// custom constants

#define 		 DBIERR_NONE  	 0				 // NO ERROR		
#define      crNOOP         0				 // No operation 
#define      crADD          1           // Add a new element.
#define      crCOPY         2           // Copy an existing element.
#define      crMODIFY       3           // Modify an element.
#define      crDROP         4           // Removes an element.
#define      dbiOPENSHARED  0           // Open shared  (Default)
#define      dbiOPENEXCL    1           // Open exclusive
#define      dbiREADWRITE   0           // Read + Write   (Default)
#define      dbiREADONLY    1           // Read only
#define      xltNONE        0           // No translation  (Physical Types)
#define      xltRECORD      1           // Record level (not supported)
#define      xltFIELD       2           // Field level transl. (Logical types)

****************
* ModiStruClass*
****************

Class modistruclass

   // Constructor
   set procedure to :dUFLP:api_structs.cc additive

   **	  Error-Handling		
   this.ErrorNumber			= 0					  // 0 = no error
   this.ErrorString			= space(128)		  //retieved with DbiGetErrorString

   **	  Table related properties   
   this.oDataBase				= null	// Database, if not specified, standard db
   this.hDB						= 0 		// handle to the above database
   this.cTableName			= ''		// Name of the table to be restructured
   this.hTable		   		= 0 		// handle to the table
   this.hCursor				= 0		// handle to a cursor, defined by the table
   this.cFieldName			= ''		// name of the field to be changed
   this.iFieldNr				= 0		// position of the field in the FLDDesc-array	 
	this.NewFieldName   		= ''
	this.NewFieldType   		= ''
	this.NewFieldLength 		= 0
	this.NewFielddecimal		= 0
	this.position				= 0
   **	  BDE-Structures
   this.sProps					= new curprops()			// curprops structure
   this.pszTableName			= new stringpointer()   // pointer to the tablename
   this.phCursor				= new handlepointer()	// bde-cursor-handle-pointer
   this.sFieldDesc			= new flddesc()			// for the field-descriptor
   this.sTblDesc				= new crtbldesc()			// for the table descriptor

   **		Methods	
   this.init			= class::init		   // to initialize the table-related props 
   this.restructure	= class::restructure // do the restructure
   this.release		= class::cleanup	   // cleanup structures

   **		make the assocarray for the fieldtype
   this.a = new AssocArray()
   this.a["PXCHAR"] 			= 0x101 
   this.a["PXNUMERIC"] 		= 0x102        // Numeric
   this.a["PXMONEY"] 		= 0x103        // Money
   this.a["PXDATE"] 			= 0x104        // Date
   this.a["PXSHORT"]			= 0x105        // Short
   this.a["PXMEMO"]			= 0x106        // Text Memo       (blob)
   this.a["PXBINARYBLOB"]	= 0x107        // Binary data     (blob)
   this.a["PXFMTMEMO"]		= 0x108        // Formatted text  (blob)
   this.a["PXOLEBLOB"]		= 0x109        // OLE object      (blob)
   this.a["PXGRAPHIC"]		= 0x10A        // Graphics object (blob)
   this.a["PXLONG"]			= 0x10B        // Long
   this.a["PXTIME"]		   = 0x10C        // Time
   this.a["PXDATETIME"]		= 0x10D        // Time Stamp
   this.a["PXBOOL"]			= 0x10E        // Logical
   this.a["PXAUTOINC"]		= 0x10F        // Auto increment (long)
   this.a["PXBYTES"]			= 0x110        // Fixed number of bytes
   this.a["PXBCD"]			= 0x111        // BCD (32 digits)
   this.a["DBCHAR"]			= 0x201        // Char string
   this.a["DBNUMERIC"]		= 0x202        // Number
   this.a["DBMEMO"]			= 0x203        // Memo          (blob)
   this.a["DBBOOL"]			= 0x204        // Logical
   this.a["DBDATE"]			= 0x205        // Date
   this.a["DBFLOAT"]			= 0x206        // Float
   this.a["DBDBLOCK"]		= 0x207        // Logical type is LOCKINFO
   this.a["DBOLE"]			= 0x208        // OLE object    (blob)
   this.a["DBBINARY"]		= 0x209        // Binary data   (blob)
   this.a["DBLONG"]			= 0x20B			 // Long
   this.a["DBTIMESTAMP"]	= 0x20C			 // Timestamp
   this.a["DBDOUBLE"]		= 0x20D			 // Double
   this.a["DBAUTOINC"]		= 0x20E			 // Autoincrement

   ** 	Externalize BDE-Apis
   extern cshort DbiGetCursorProps(chandle,cptr) idapi32
   extern cshort DbiGetFieldDescs(chandle,cptr) idapi32
   extern cshort DbiOpenTable(chandle,clong,cptr,cptr,cptr,cshort,cshort,cshort, ;
                              cshort,clogical,cptr,cptr) idapi32
   extern cshort DbiCloseCursor(cptr) idapi32
   extern cshort DbiDoRestructure(chandle,cshort,cptr,cptr,cptr,cptr,clogical) idapi32 
   extern cshort DbiGetErrorString(cshort,cstring) idapi32

   // End of constructor


   // Methods

   /* 
      The init-method is called by the other methods , to make sure, 
      that after a change in the structure of the record,
      the fielddescriptors are up to date
   */
   function init()
      local q

      if this.oDataBase = null		
         this.hdb = _app.databases[1].handle		// standard database
      else
         this.hdb = this.oDataBase.handle			// user defined database
      endif

      if empty(this.cTableName)
         this.ErrorNumber = 1
         this.ErrorString = "Missing Tablename"
         return
      endif

      // get the pointer to the tablename-string
      q=val(this.psztablename.pstring(this.ctablename))	
      // now try to open this table exclusively with translation-mode physical
      this.ErrorNumber = dbiopentable(this.hdb,q,null,null,null,0,0, ;
                         dbiOPENEXCL,xltnone,false,null,this.phcursor)	
      if this.Errornumber # DBIERR_NONE 
         dbigeterrorstring(this.errornumber,this.errorstring)
         return
      endif

      // get the handle to the cursor associated with the open table
      this.hcursor=this.phcursor.hnd()

      // get the properties of this cursor
      this.Errornumber = dbigetcursorprops(this.hcursor,this.sProps)
      if this.Errornumber # DBIERR_NONE 
         dbigeterrorstring(this.errornumber,this.errorstring)
         return
      endif

      // Only dBase or Paradox tables can be restructured
      q=this.sprops.sztabletype()
      if (q # szPARADOX) and (q # szDBASE)
         this.errorstring = "Only dBase or Paradox tables may be restructured"
         this.errornumber = 2
         return
      endif

      if q=szparadox	
         this.cshort="PX"
      else 
         this.cshort="DB"
      endif

      // prepare an array of field descriptor structures and get the fieldinfos
      this.aFldDesc=new struct_base(62*this.sprops.ifields())    
      this.errornumber = dbigetfielddescs(this.hcursor,this.aFldDesc)
      if this.Errornumber # DBIERR_NONE 
         dbigeterrorstring(this.errornumber,this.errorstring)
         return
      endif
      // the cursor must be closed before any restructuring
      this.errornumber = dbiclosecursor(this.phcursor)
   return

   /* 
      The restructure-method can be called 
      after having set the new field properties
   */
   function restructure()
      local i,ifield,opdesc,n

      if len(this.cFieldName) = 0
         this.errorstring = "You must specify a fieldname"
         this.errornumber=3
         return
      endif

		this.init()
		if this.Errornumber # DBIERR_NONE 
	 		dbigeterrorstring(this.errornumber,this.errorstring)
		   return
		endif

      // get the number of fields in the table
      ifield=this.sprops.ifields()
      this.errornumber=4

      // search the fields of the table to find the one stated in cFieldName
      for i=1 to ifield
         this.sFieldDesc.string=substr(this.aFldDesc.string,(i-1)*31+1,31)
         if upper(ltrim(rtrim(this.sfielddesc.szname()))) ==  ;
            upper(ltrim(rtrim(this.cfieldname)))
            this.ifieldnr = i
            this.errornumber=0
            i=ifield+1
         endif
      endfor
      if this.Errornumber # DBIERR_NONE 
         this.errorstring = "The fieldname does not exist"
         return
      endif

      // Prepare the structure for the operation descriptors
      opdesc=new struct_base(ifield*4)  
      i=this.ifieldnr-1
      opdesc.setlong(i*4,crModify)

      if len(this.NewFieldName) # 0
         ********   MEOB 2012/12/01
         cNull = replicate(chr(0),32)
         this.aflddesc.setansistring(i*62+2,31,cNull) // blank out original fieldname
         ********
         this.aflddesc.setansistring(i*62+2,31,this.NewFieldName) // set new fieldname
      endif        

      if len(this.NewFieldType) # 0
         this.NewFieldType=this.cshort+upper(ltrim(rtrim(this.Newfieldtype)))
         if not this.a.iskey(this.newfieldtype)
            this.errornumber = 4
            this.errorstring = "This Fieldtype is not allowed"
            return
         endif
         this.aflddesc.setword(i*62+34,this.a[this.newfieldtype])	// set new type
      endif

      if this.NewFieldlength # 0
         this.aflddesc.setword(i*62+38,this.Newfieldlength)		// set new length
      endif

      if this.NewFielddecimal #0 
         this.aflddesc.setword(i*62+40,this.NewFielddecimal)     // set new decimal
      endif

      // not sure if this is necessary, but it occurs in the bde-help,
      for n= 0 to ifield-1
         this.aflddesc.setword(n*62,n+1)
      endfor

  
      * Now fill the table descriptor structure
      this.stbldesc.sztblname(this.ctablename)
      this.stbldesc.sztbltype(this.sprops.sztabletype())
      this.stbldesc.ifldcount(ifield)
      this.stbldesc.pcroptype(opdesc)
      this.stbldesc.pflddesc(this.aflddesc)

      * leave out the other members, they are not used for field-restructure
      * then call the DbiDoRestructure BDE-API-Function
      this.errornumber = dbidorestructure(this.hdb,1,this.stbldesc,null, ;
                                          null,null,false)
      if this.Errornumber # DBIERR_NONE 
         dbigeterrorstring(this.errornumber,this.errorstring)
         return
      endif

      // reset the Field-specifications
		this.NewFieldName   = ''
		this.NewFieldType   = ''
		this.NewFieldLength = 0
		this.NewFielddecimal= 0
      opdesc.release()
      this.aflddesc.release()
   return

	function addfield()

		/* fill the properties with the appropriate values  example
		this.NewFieldName   = 'Nickname'
		this.NewFieldType   = 'character'
		this.NewFieldLength = 10
		this.NewFielddecimal= 0
	
		*also fill the position-property (that is the position, where the new field 
		*should be inserted
		
		this.position = 2 (add field at second pos)  ( 0 means add at the end)
	
		*call the method
	
		this.addfield()
		*/
	
		local i,ifield,opdesc,n,ne

		if len(this.NewFieldName) = 0
			this.errorstring = "You must specify a fieldname"
	   	this.errornumber=3
		   return
		endif
		this.init()
		if this.Errornumber # DBIERR_NONE 
	 		dbigeterrorstring(this.errornumber,this.errorstring)
		   return
		endif
		// get the number of fields in the table and add 1
		ifield=this.sprops.ifields()+1
		// enlarge the aflddesc-structure to hold the new field at the desired position
		if this.position = 0 or this.position> ifield
			this.position=ifield
		   this.aflddesc.string = this.aflddesc.string + replicate(chr(0),31)
		else
			if this.position=1
		   	this.aflddesc.string=replicate(chr(0),31)+this.aflddesc.string
		   else
				this.aflddesc.string=left(this.aflddesc.string,(this.position-1)*31)+ ;
            						  replicate(chr(0),31)+ ;
	   						        right(this.aflddesc.string,(ifield	-this.position)*31)
			endif 	
		endif
	
		// Prepare the structure for the operation descriptors
		opdesc=new struct_base(ifield*4)  
		i=this.position-1
		opdesc.setlong(i*4,crAdd)
		this.aflddesc.setansistring(i*62+2,31,this.NewFieldName) // set new fieldname
		this.NewFieldType=this.cshort+upper(ltrim(rtrim(this.Newfieldtype)))
		if not this.a.iskey(this.newfieldtype)
		  	this.errornumber = 4
		   this.errorstring = "This Fieldtype is not allowed"
		   return
		endif
		this.aflddesc.setword(i*62+34,this.a[this.newfieldtype])	// set new type
		this.aflddesc.setword(i*62+38,this.Newfieldlength)		// set new length
		this.aflddesc.setword(i*62+40,this.NewFielddecimal)     // set new decimal
		ne=1
		for n= 0 to ifield-1
      	if n=i
         	this.aflddesc.setword(n*62,0)
         else
	 			this.aflddesc.setword(n*62,ne)
            ne+=1
         endif
  		endfor
		  
		* Now fill the table descriptor structure
		
		this.stbldesc.sztblname(this.ctablename)
		this.stbldesc.sztbltype(this.sprops.sztabletype())
		this.stbldesc.ifldcount(ifield)
		this.stbldesc.pcroptype(opdesc)
		this.stbldesc.pflddesc(this.aflddesc)
		this.errornumber = dbidorestructure(this.hdb,1,this.stbldesc,null, ;
														null,null,false)
		if this.Errornumber # DBIERR_NONE 
			dbigeterrorstring(this.errornumber,this.errorstring)
		   return
		endif

		// reset the Field-specifications
		this.NewFieldName   = ''
		this.NewFieldType   = ''
		this.NewFieldLength = 0
		this.NewFielddecimal= 0
		opdesc.release()
		this.aflddesc.release()
   return

   function DeleteField()

		/* set the fieldname that is to be deleted

      this.cFieldName   = 'Nickname'
	
		*call the method
	
		this.DeleteField()
		*/
	
		local i,ifield,opdesc,n,ne

		if len(this.cFieldName) = 0
			this.errorstring = "You must specify a fieldname"
	   	this.errornumber=3
		   return
		endif
		this.init()
		if this.Errornumber # DBIERR_NONE 
	 		dbigeterrorstring(this.errornumber,this.errorstring)
		   return
		endif

      // get the number of fields in the table
      ifield=this.sprops.ifields()
      this.errornumber=4

      // search the fields of the table to find the one stated in cFieldName
      for i=1 to ifield
         this.sFieldDesc.string=substr(this.aFldDesc.string,(i-1)*31+1,31)
         if upper(ltrim(rtrim(this.sfielddesc.szname()))) ==  ;
            upper(ltrim(rtrim(this.cfieldname)))
            this.position = i
            this.errornumber=0
            i=ifield+1
         endif
      endfor
      if this.Errornumber # DBIERR_NONE 
         this.errorstring = "The fieldname does not exist"
         return
      endif

      // Prepare the structure for the operation descriptors

      opdesc=new struct_base((ifield-1)*4)  

		// change the aflddesc-structure 

		if this.position= ifield
		   this.aflddesc.string = left(this.aflddesc.string , (ifield-1)*31)
		else
			if this.position=1
		   	this.aflddesc.string=right(this.aflddesc.string,(ifield-1)*31)
		   else
				this.aflddesc.string=left(this.aflddesc.string,(this.position-1)*31)+ ;
	   						        right(this.aflddesc.string,(ifield-this.position)*31)
			endif 	
		endif
      ne=0
		for n= 0 to ifield-1
      	if n # this.position - 1
	 			this.aflddesc.setword(ne*62,n+1)
            ne+=1
         endif
  		endfor
      this.stbldesc.sztblname(this.ctablename)
      this.stbldesc.sztbltype(this.sprops.sztabletype())
      this.stbldesc.ifldcount(ifield-1)
      this.stbldesc.pcroptype(opdesc)
      this.stbldesc.pflddesc(this.aflddesc)
      this.errornumber = dbidorestructure(this.hdb,1,this.stbldesc,null, ;
                                          null,null,false)
      if this.Errornumber # DBIERR_NONE 
         dbigeterrorstring(this.errornumber,this.errorstring)
         return
      endif
	return
   	

   function cleanup
      if type( "this.sflddesc") # "U"
         this.sflddesc.release()
      endif
      if type( "this.phcursor") # "U"
         this.phcursor.release()
      endif
      if type( "this.sprops") # "U"
         this.sprops.release()
      endif
      if type( "this.psztablename") # "U"
         this.psztablename.release()
      endif
*      if type( "this.stbldesc") # "U"
 *        this.stbldesc.release()
  *    endif
   return

   function getdblocklength()

   /* This method is added, because there is no way in oodml to read the 
   	length of a dbaselockfield.

      Simply call the function. Return value is 0 on error or length of field.

      ilength=t.getdblocklength()
    */

   	local i,locklength,ifield

		this.init()
		if this.Errornumber # DBIERR_NONE 
	 		dbigeterrorstring(this.errornumber,this.errorstring)
		   return 0
		endif
		
      // get the number of fields in the table
      ifield=this.sprops.ifields()
      this.errornumber=4

      // search the fields of the table to find a _DBASELOCK
      for i=1 to ifield
         this.sFieldDesc.string=substr(this.aFldDesc.string,(i-1)*31+1,31)
         if this.sfielddesc.szname() == "_DBASELOCK"
            this.ifieldnr = i
            this.errornumber=0
            i=ifield+1
         endif
      endfor
      if this.Errornumber # DBIERR_NONE 
         this.errorstring = "There is no Lock-Field"
         return 0
      endif

		locklength=this.aflddesc.getword((this.ifieldnr-1)*62+38)	

      this.aflddesc.release()
   return locklength

Endclass				// end of class ModiStruClass






