// Metric: 0 - Chars or 6 - Pixels only.
/*
    MOVER.CC
    Author: Ken Mayer
    Date..: 04/10/1998
    Notes.: This is custom container with two listboxes and 
            some buttons and code to handle moving items 
            between listboxes. This is meant to be a relatively
            standard control, based on what can be seen
            in many applications. It is pure dBL, and should
            work in all versions of dBASE that include
            the Drag and Drop feature (dB2K 0.1 and later,
            including dBASE SE, dBASE Plus, etc.)

    Updated: May 31, 2000 -- Eric Bachleitner
             Eric updated this so that mover is now
             self-contained -- i.e., you can have multiple
             mover controls on the same form, and more ...

             August 5, 2002 -- Ken Mayer
             Added drag and drop functionality (to drag/
             drop items from one listbox to the other)
             This meant adding some events, and modifying
             the onClick events to handle a parameter
             for moving items by dragging them.

             Updated the comments below to help reflect 
             what needs to be done more clearly -- after
             Eric's changes, I forgot to update the details
             in these comments, so that one could easily
             get confused ...

             October 23, 2002 -- Ken Mayer
             Fixed a couple minor bugs pointed out by Bruce
             Beacham. Made variables local to individual routines ...

             August 29, 2003 -- KJM
             Made it possible to set the DestArray from outside
             the control to some array.

             September 19, 2003 -- KJM
             Modified so that the drag within a listbox,
             rather than moving something from the other
             listbox to the current one, moves from the current
             to the other. This is not quite optimal, but 
             I'm not sure what else to do ... the problem
             is that if you drag within the listbox (up/down)
             you can't do it with the listbox as it sits now
             without a huge amount of work. 

             January 16, 2004 -- KJM
             Fixed the drag and drop. I was calling the code backward
             when dropping something from one listbox to the other.
             Duh. It's been fixed, and appears to work like a charm.

             I also fixed the problem with dropping something onto
             the source when dragging (i.e., when on the left listbox,
             if you drag an item, and then drop it, the 'drop' event
             was firing, which was odd, to say the least; I have found
             that I avoid that occurring, which means items are not
             being moved when they shouldn't be).

    *** WARNING ***
    This control uses a container with the metric set to
    Chars (0) -- if your form uses a different metric this
    can cause some problems (there's a known problem in
    dBASE, and a bug report exists for this issue). The 
    current solution: set the form's metric to Chars, 
    drop this control on it, then change the form's metric
    to what you normally use ...
    *** END OF WARNING ***

    Usage:

      First, place the custom container onto a form:
           set procedure to :duflp:mover.cc additive
           // in the designer drop a copy of MOVERCONTAINER
           // onto a form, add whatever other controls
           // you wish ...

      When working with the form that has an instance
      of MOVER2CONTAINER, you must assign the source 
      array to the Container's source array -- this needs
      to be a single-dimension (one column) array:

           set procedure to mover.wfm additive
           fMover = new moverform()
           // 'Mover2Container1' is the default
           // name given this container when
           // dropped onto a form
           fMover.Mover2Container1.SourceArray = myarrayreference
           // optional, set the destination array to some
           // array, which allows you to set some default
           // selections:
           fMover.Mover2Container1.DestArray = anotherarrayreference
           fMover.Mover2Container1.reassertArrays()
           fMover.ReadModal()

      That should be all -- the container will make a copy
      of your original source for use in the form, so
      that we don't destroy your original array.

      OPTIONAL:
         These two text controls are over the two
         listboxes -- you may wish to change the 
         text:

         fMover.mover2Container1.SourceTitle.text = "Something"
         fMover.mover2Container1.DestTitle.text   = "Somethingelse"

      If you want to do something else as well as move items 
      from one list to another:
         To handle this, you need to OVERRIDE one of the events.
         The most likely event to override is the ReassertArrays
         event, because this one actually changes the values in the
         arrays and re-assigns the lists for the listboxes.

         For example, if you are working with files, you might
         wish to add code that totals up the number of bytes
         of the files selected. 

         Overriding the event is as simple as adding code in your
         form to perform the override. Do this in the form's
         onOpen event:

         form.mover2container1.reassertArrays = class::reassertArrays

         Then add a new function (event) in the form, before the
         statement "ENDCLASS" at the end of the file (without the 
         comment operators):

         // function reassertArrays
         //   // Call original code, or it won't execute:
         //   mover2container::reassertArrays()
         //
         //   // do your own code here
         //
         // return

      =========================
      TO GET DESTINATION ARRAY:
      =========================
      To get the data from the destination array, use
      code similar to that below:

          if fMover.Mover2Container1.destArray[1] # "<Empty>"
             myNewArray = new Array()
             for i = 1 to fMover.Mover2Container1.destArray.size
                 mynewarray.add( fMover.Mover2Container1.destArray[i] )
             next
          else
             ? "Nothing to process"
          endif


      To see an example, run :dUFLP:TestMover.wfm
*/

class MOVER2CONTAINER( oFormObj ) of CONTAINER( oFormObj ) custom
   with ( this )
      transparent := true
      width       := 40
      height      := 9.5
      borderStyle := 10	// Etched Out
      onOpen      := CLASS::OpenIt
      metric      := 0 // Chars
   endwith

   this.SOURCETITLE = new TEXT(this)
   with (this.SOURCETITLE)
      height = 0.8
      left = 2
      top = 0.44
      width = 12
      border = true
      colorNormal = "blue"
      alignVertical = 1	// Middle
      alignHorizontal = 1	// Center
      fontBold = true
      text = "Source"
      borderStyle = 1	// Raised
   endwith

   this.DESTTITLE = new TEXT(this)
   with (this.DESTTITLE)
      height = 0.8
      left = 25.5556
      top = 0.44
      width = 12
      border = true
      colorNormal = "blue"
      alignVertical = 1	// Middle
      alignHorizontal = 1	// Center
      fontBold = true
      text = "Destination"
      borderStyle = 1	// Raised
   endwith

   this.SOURCELISTBOX = new LISTBOX(this)
   with (this.SOURCELISTBOX)
      allowDrop = true
      dragEffect = 1 // Copy
      onDrop = class::SourceDrop
      onLeftMouseDown = class::MoveRight
      onLeftDblClick = {;This.parent.MoveCurrentRightButton_onClick()}
      height = 7.6
      left = 1.2222
      top = 1.44
      width = 13.4444
      id = 101
      colorHighLight = "white/blue"
      dataSource = 'ARRAY {"Source Array"}'
      multiple = true
   endwith

   this.DESTLISTBOX = new LISTBOX(this)
   with (this.DESTLISTBOX)
      allowDrop = true
      dragEffect = 1 // Copy
      onDrop = class::DestDrop
      onLeftMouseDown = class::MoveLeft
      onLeftDblClick = {;This.parent.MoveCurrentLeftButton_onClick()}
      height = 7.6
      left = 25
      top = 1.44
      width = 13.4444
      id = 101
      colorHighLight = "white/blue"
      dataSource = 'ARRAY {"Dest. Array"}'
      multiple = true
   endwith

   this.BTNCONTAINER = new CONTAINER(this)
   with (this.BTNCONTAINER)
      transparent = true
      left = 15.7778
      top = 1.76
      width = 8
      height = 6.48
      borderStyle = 3	// None
   endwith

   this.BTNCONTAINER.MOVECURRENTRIGHTBUTTON = new PUSHBUTTON(this.BTNCONTAINER)
   with (this.BTNCONTAINER.MOVECURRENTRIGHTBUTTON)
      onClick = {;this.parent.parent.MOVECURRENTRIGHTBUTTON_ONCLICK()}
      height = 1.08
      left = 1.1111
      top = 0.6
      width = 6.1111
      text = ""
      speedBar = true
      upBitmap = "RESOURCE #854"
      speedTip = "Move Current Item(s) to Destination"
      pageno = 0
      value = false
   endwith

   this.BTNCONTAINER.MOVEALLRIGHTBUTTON = new PUSHBUTTON(this.BTNCONTAINER)
   with (this.BTNCONTAINER.MOVEALLRIGHTBUTTON)
      onClick = {;this.parent.parent.MOVEALLRIGHTBUTTON_ONCLICK()}
      height = 1.08
      left = 1.1111
      top = 1.72
      width = 6.1111
      text = ""
      speedBar = true
      upBitmap = "RESOURCE #855"
      speedTip = "Move All Item to Destination"
      pageno = 0
      value = false
   endwith

   this.BTNCONTAINER.MOVECURRENTLEFTBUTTON = new PUSHBUTTON(this.BTNCONTAINER)
   with (this.BTNCONTAINER.MOVECURRENTLEFTBUTTON)
      onClick = {;this.parent.parent.MOVECURRENTLEFTBUTTON_ONCLICK()}
      height = 1.08
      left = 1.1111
      top = 3.4
      width = 6.1111
      text = ""
      speedBar = true
      upBitmap = "RESOURCE #853"
      speedTip = "Move Current Item(s) to Source"
      pageno = 0
      value = false
   endwith

   this.BTNCONTAINER.MOVEALLLEFTBUTTON = new PUSHBUTTON(this.BTNCONTAINER)
   with (this.BTNCONTAINER.MOVEALLLEFTBUTTON)
      onClick = {;this.parent.parent.MOVEALLLEFTBUTTON_ONCLICK()}
      height = 1.08
      left = 1.1111
      top = 4.48
      width = 6.1111
      text = ""
      speedBar = true
      upBitmap = "RESOURCE #852"
      speedTip = "Move All Item to Source"
      pageno = 0
      value = false
   endwith

   function onDesignOpen( lFromPalette )
      if lFromPalette and form.Metric == 6 // Pixels
         this.height                                     := 209
         this.width                                      := 280
         this.DestListBox.height                         := 167
         this.DestListBox.left                           := 175
         this.DestListBox.top                            :=  32
         this.DestListBox.width                          :=  94
         this.SourceListBox.height                       := 167
         this.SourceListBox.left                         :=   9
         this.SourceListBox.top                          :=  32
         this.SourceListBox.width                        :=  94
         this.SourceTitle.height                         :=  18
         this.SourceTitle.left                           :=  14
         this.SourceTitle.top                            :=  10
         this.SourceTitle.width                          :=  84
         this.DestTitle.height                           :=  18
         this.DestTitle.left                             := 179
         this.DestTitle.top                              :=  10
         this.DestTitle.width                            :=  84
         this.BtnContainer.height                        := 143
         this.BtnContainer.left                          := 110
         this.BtnContainer.top                           :=  39
         this.BtnContainer.width                         :=  56
         this.BtnContainer.MoveAllLeftButton.height      :=  24
         this.BtnContainer.MoveAllLeftButton.left        :=   8
         this.BtnContainer.MoveAllLeftButton.top         :=  99
         this.BtnContainer.MoveAllLeftButton.width       :=  43
         this.BtnContainer.MoveAllRightButton.height     :=  24
         this.BtnContainer.MoveAllRightButton.left       :=   8
         this.BtnContainer.MoveAllRightButton.top        :=  38
         this.BtnContainer.MoveAllRightButton.width      :=  43
         this.BtnContainer.MoveCurrentLeftButton.height  :=  24
         this.BtnContainer.MoveCurrentLeftButton.left    :=   8
         this.BtnContainer.MoveCurrentLeftButton.top     :=  75
         this.BtnContainer.MoveCurrentLeftButton.width   :=  43
         this.BtnContainer.MoveCurrentRightButton.height :=  24
         this.BtnContainer.MoveCurrentRightButton.left   :=   8
         this.BtnContainer.MoveCurrentRightButton.top    :=  13
         this.BtnContainer.MoveCurrentRightButton.width  :=  43
      endif
   return

   function MOVEALLLEFTBUTTON_onClick
      /* 
         ----------------------------------------
         Move all items in destination back
         to source:
         ----------------------------------------
      */
      local i

      // First, check to see if we're working with an
      // "empty" array ...:
      if aLen( This.destArray ) == 1 and ;
         This.destArray[1] == "<Empty>"
         return
      endif

      // Copy the elements in the destination array
      // back to the source array -- fun part is
      // that there may be some elements already
      // there ...:
      if This.copySource[1] == "<Empty>"
         This.copySource = new Array()
      endif
      for i = 1 to aLen( This.destArray )
          This.copySource.add( This.destArray[i] )
      next
      this.ResetDestination() // it's now empty
      this.ReassertArrays()   // show 'em ...

   return

   function MOVEALLRIGHTBUTTON_onClick
      /* 
         ----------------------------------------
         Move all items in source to destination:
         ----------------------------------------
      */
      local i

      // First, check to see if we're working with an
      // "empty" array ...:
      if aLen( This.copySource ) == 1 and ;
         This.copySource[1] == "<Empty>"
         return
      endif

      // Copy the elements in the source array
      // to the destination array -- fun part is
      // that there may be some elements already
      // there ...:
      if This.destArray[1] == "<Empty>"
         This.destArray = new Array()
      endif
      for i = 1 to aLen( This.copySource )
          This.destArray.add( This.copySource[i] )
      next
      this.EmptySource() // it's now empty
      this.ReassertArrays()   // show 'em ...

   return

   function MOVECURRENTLEFTBUTTON_onClick( cItem )
        /* 
         ----------------------------------------
         Move current item(s) in destination back
         to source (this is called both from
         a left-double-click on the listbox,
         and from the pushbutton named above):

         August, 2002, added parameter, so that
         this code can be called from the 
         drag and drop methods ...
         ----------------------------------------
      */
      private oThis, cObjRef, cSize, cArray, i

      if pCount() == 1 and not empty( cItem )
         // we need to modify this a bit:
         oThis = this.parent
      else
         oThis = this
      endif

      // First, check to see if we're working with the
      // "empty" array ...:
      if aLen( oThis.destArray ) == 1 and ;
         oThis.destArray[1] == "<Empty>"
         return
      endif

      // Now for the tricky part -- moving just the
      // element(s) selected:
      if oThis.copySource[1] == "<Empty>"
         oThis.copySource = new array()
      endif

      // Need an object reference as a string
      // for this, since we have to build it
      // for the macro
      cObjRef = "oThis.DestListBox"
      cSize = cObjRef+".selected().size"
      cArray = cObjRef+".selected()[i]"
      for i = 1 to &cSize.
         oThis.copySource.add( &cArray. )

          // delete this element from the listbox's array:
          nRow = oThis.destArray.scan( &cArray. )
          oThis.destArray.delete( nRow )
          oThis.destArray.size--
      next

      // if destination is now empty:
      if oThis.destArray.size = 0
         oThis.ResetDestination()
      endif
      oThis.ReassertArrays()

   return

   function MOVECURRENTRIGHTBUTTON_onClick( cItem )
        /* 
         ----------------------------------------
         Move current item(s) in source
         to destination (this is called both from
         a left-double-click on the listbox,
         and from the pushbutton named above):

         August, 2002, added parameter, so that
         this code can be called from the 
         drag and drop methods ...
         ----------------------------------------
      */
      private oThis, cObjRef, cSize, cArray, i

      if pCount() == 1 and not empty( cItem )
         // we need to modify this a bit:
         oThis = this.parent
      else
         oThis = this
      endif

      // First, check to see if we're working with the
      // "empty" array ...:
      if aLen( oThis.CopySource ) == 1 and ;
         oThis.CopySource[1] == "<Empty>"
         return
      endif

      // Now for the tricky part -- moving just the
      // element(s) selected:
      if oThis.destArray[1] == "<Empty>"
         oThis.destArray = new array()
      endif    

      // Need an object reference as a string
      // for this, since we have to build it
      // for the macro
      cObjRef	= "oThis.SourceListBox"
      cSize  = cObjRef +".selected().size"
      cArray = cObjRef +".selected()[i]"
      for i = 1 to &cSize.
          oThis.destArray.add( &cArray. )

          // delete this element from the listbox's array:
          nRow = oThis.copysource.scan( &cArray. )
          oThis.copysource.delete( nRow )
          oThis.copysource.size--
      next

      // if source is now empty:
      if oThis.CopySource.size = 0
         oThis.EmptySource()
      endif

      oThis.ReassertArrays()
   return

   function MoveLeft
      /*
         Drag item ... this initiates the drag operation
         called from the DestListBox's onLeftMouseDown event
      */
 		this.parent.DragSource := this.name
      this.drag( this.name, this.value, "" )

   function MoveRight
      /*
         Drag item ... this initiates the drag operation
         called from the SourceListBox's onLeftMouseDown event
      */
		this.parent.DragSource := this.name
      this.drag( this.name, this.value, "" )

   function SourceDrop(nLeft, nTop, cType, cName)
      /*
         onDrop event for the Source Listbox -- we need
         to call the appropriate code for moving this
         item ... we're moving it from the right to the left
      */
      if this.parent.DragSource == this.name
         return
      endif
      class::MoveCurrentLeftButton_onClick( cName )
      this.parent.DragSource = ""

   function DestDrop(nLeft, nTop, cType, cName)
      /*
         onDrop event for the Destination Listbox -- we need
         to call the appropriate code for moving this
         item ... we're moving it from the left to the right:
      */
      if this.parent.DragSource == this.name
         return
      endif
      class::MoveCurrentRightButton_onClick( cName )
      this.parent.DragSource = ""

   PROCEDURE OpenIt
      // When the container is opened:
      local i
      this.DragSource = "" // init

      // assign a simple array to start
      // with to the destination listbox:
      if type( "this.DestArray" ) # "A" or ;
         type( "this.DestArray" ) == "U"
         this.DestArray = new Array()
         this.DestArray.add( "<Empty>" )
      endif
      this.DestListBox.datasource = "ARRAY this.DestArray"

      // if no source array exists:
      if type( "this.SourceArray" ) == "U" or ;
         type( "this.SourceArray" ) # "A"
         msgbox( "Please assign an array to the "+;
                 "mover container's SourceArray property!",;
                 "Mover Error!", 16 )
         form.close()
         return
      endif

      // Copy the source array:
      this.CopySource = new array()
      for i = 1 to alen( this.SourceArray )
          this.CopySource.add( this.SourceArray[ i ] )
      next

      // assign to source listbox:
      this.SourceListBox.datasource :=;
                 "ARRAY this.CopySource"

   procedure resetDestination
      // reset destination array to empty:
      if this.baseClassName == "LISTBOX"
         oThis = this.parent
      else
         oThis = this
      endif
      oThis.DestArray = new Array()
      oThis.DestArray.add( "<Empty>" )

      oThis.DestListBox.datasource := "ARRAY oThis.DestArray"

   procedure emptySource
      // if the source is empty:
      if this.baseClassName == "LISTBOX"
         oThis = this.parent
      else
         oThis = this
      endif
      oThis.copySource = new Array()
      oThis.copySource.add( "<Empty>" )
      oThis.SourceListBox.datasource := "ARRAY oThis.copySource"

   procedure reassertArrays
      // Ensure that the listboxes are showing the correct
      // values in the arrays
      if this.baseClassName == "LISTBOX"
         oThis = this.parent
      else
         oThis = this
      endif
      othis.DestListBox.DataSource := "ARRAY oThis.DestArray"
		oThis.SourceListBox.DataSource := "ARRAY oThis.copySource"

endclass // MOVERCONTAINER

// -----------------------
// End of File: MOVER.CC
// -----------------------

